# -*- coding: utf-8 -*-


# Built-in
import copy


# Common
import numpy as np


# local
from ._class01_Mesh2D import Mesh2D as Previous
from . import _class01_select as _select
from . import _class02_checks as _checks
from . import _class02_compute as _compute
from . import _class02_contours as _contours
from . import _class01_rect_cropping as _cropping
from . import _class02_interpolate as _interpolate
from . import _class02_interpolate_all as _interpolate_all
from . import _class02_operators as _operators
from . import _class02_plot_as_profile2d as _plot_as_profile2d
from . import _class02_plot_as_profile2d_compare as _plot_as_profile2d_compare


__all__ = ['BSplines2D']


# #############################################################################
# #############################################################################
#
# #############################################################################


class BSplines2D(Previous):

    _ddef = copy.deepcopy(Previous._ddef)
    _ddef['params']['ddata'].update({
        'bsplines': {'cls': (tuple, str), 'def': ''},
    })
    _ddef['params']['dobj'] = None
    _ddef['params']['dref'] = None

    _dshow = dict(Previous._dshow)
    _dshow['data'].append('bsplines')

    # -----------------
    # bsplines
    # ------------------

    def add_bsplines(self, key=None, deg=None):
        """ Add bspline basis functions on the chosen mesh """

        # --------------
        # check inputs

        keym, keybs, deg = _checks._mesh_bsplines(
            key=key,
            lkeys=list(self.dobj[self._which_mesh].keys()),
            deg=deg,
        )

        # ------------
        # get bsplines

        nd = self.dobj[self._which_mesh][keym]['nd']
        mtype = self.dobj[self._which_mesh][keym]['type']
        if nd == '1d':
            dref, ddata, dobj = _compute._mesh1d_bsplines(
                coll=self, keym=keym, keybs=keybs, deg=deg,
            )
        elif mtype == 'rect':
            dref, ddata, dobj = _compute._mesh2DRect_bsplines(
                coll=self, keym=keym, keybs=keybs, deg=deg,
            )
        elif mtype == 'tri':
            dref, ddata, dobj = _compute._mesh2DTri_bsplines(
                coll=self, keym=keym, keybs=keybs, deg=deg,
            )
        else:
            dref, ddata, dobj = _compute._mesh2Dpolar_bsplines(
                coll=self, keym=keym, keybs=keybs, deg=deg, # angle=angle,
            )

        # --------------
        # update dict and crop if relevant

        self.update(dobj=dobj, ddata=ddata, dref=dref)
        if mtype == 'rect':
            _cropping.add_cropbs_from_crop(
                coll=self,
                keybs=keybs,
                keym=keym,
            )

    # -----------------
    # remove_bsplines
    # ------------------

    def remove_bsplines(self, key=None, propagate=None):
        return _checks.remove_bsplines(coll=self, key=key, propagate=propagate)

    # -----------------
    # add_data
    # ------------------

    def update(
        self,
        dobj=None,
        ddata=None,
        dref=None,
        harmonize=None,
    ):
        " Overload datastock update() method "

        # if ddata => check ref for bsplines
        if ddata is not None:
            for k0, v0 in ddata.items():
                (
                    ddata[k0]['ref'], ddata[k0]['data'],
                ) = _checks.add_data_meshbsplines_ref(
                    coll=self,
                    ref=v0['ref'],
                    data=v0['data'],
                )

        # update
        super().update(
            dobj=dobj,
            ddata=ddata,
            dref=dref,
            harmonize=harmonize,
        )

        # assign bsplines
        _checks._set_data_bsplines(coll=self)

    # -----------------
    # crop
    # ------------------

    def crop(self, key=None, crop=None, thresh_in=None, remove_isolated=None):
        """ Crop a mesh / bspline

        Uses:
            - a mask of bool for each mesh elements
            - a 2d (R, Z) closed polygon

        If applied on a bspline, cropping is double-checked to make sure
        all remaining bsplines have full support domain
        """
        super().crop(
            key=key,
            crop=crop,
            thresh_in=thresh_in,
            remove_isolated=remove_isolated,
        )

        # also crop bsplines
        for k0 in self.dobj.get(self._which_bsplines, {}).keys():
            if self.dobj[self._which_bsplines][k0][self._which_mesh] == key:
                _cropping.add_cropbs_from_crop(coll=self, keybs=k0, keym=key)

    # -----------------
    # get data subset
    # ------------------

    def get_dict_bsplines(self):
        """ Return dict of profiles2d with associated bsplines as values """

        # dict of profiles2d
        dbs, dref = {}, {}
        for k0, v0 in self._ddata.items():
            if v0[self._which_bsplines] not in [None, ''] and 'crop' not in k0:
                dbs[k0] = {
                    k1: [
                        v0['ref'].index(rr)
                        for rr in self.dobj[self._which_bsplines][k1]['ref']
                    ]
                    for k1 in v0[self._which_bsplines]
                }

                lax = np.concatenate(tuple([v1 for v1 in dbs[k0].values()]))
                dref[k0] = [
                    rr for ii, rr in enumerate(v0['ref'])
                    if ii not in lax
                ]

        return dbs, dref

    # -----------------
    # indices
    # ------------------

    def select_bsplines(
        self,
        key=None,
        ind=None,
        returnas=None,
        return_cents=None,
        return_knots=None,
        crop=None,
    ):
        """ Return indices or values of selected knots / cent

        Can be used to convert tuple (R, Z) indices to flat (RZ,) indices
        Can return values instead of indices
        Can return indices / values of neighbourgs

        """
        return _select._select_bsplines(
            coll=self,
            key=key,
            ind=ind,
            returnas=returnas,
            return_cents=return_cents,
            return_knots=return_knots,
            crop=crop,
        )

    # -----------------
    # utils
    # ------------------

    def get_profiles2d(self):
        """ Return dict of data relying on 2d bsplines """

        return _compute._get_profiles2d(self)

    def extract(self, keys=None, vectors=None):
        """ Extract some selected data and return as new instance

        Includes:
            - all desired data keys
            - all relevant ref
            - all associated monotonous vectors (optional)
            - all relevant bsplines and meshes

        """

        coll2 = super().extract(keys=keys, vectors=vectors)
        return _compute.extract(self, coll2=coll2, vectors=vectors)

    # -----------------
    # Integration operators
    # ------------------

    def add_bsplines_operator(
        self,
        key=None,
        operator=None,
        geometry=None,
        crop=None,
        # store vs return
        store=None,
        returnas=None,
        return_param=None,
        # specific to deg = 0
        centered=None,
        # to return gradR, gradZ, for D1N2 deg 0, for tomotok
        returnas_element=None,
    ):
        """ Get a matrix operator to compute an integral

        operator specifies the integrand:
            - 'D0N1': integral of the value
            - 'D0N2': integral of the squared value
            - 'D1N2': integral of the squared gradient
            - 'D2N2': integral of the squared laplacian

        geometry specifies in which geometry:
            - 'linear': linear geometry (cross-section = surface)
            - 'toroidal': toroildal geometry (cross-section = volumic slice)

        """

        return _operators.get_bsplines_operator(
            coll=self,
            key=key,
            operator=operator,
            geometry=geometry,
            crop=crop,
            store=store,
            returnas=returnas,
            return_param=return_param,
            # specific to deg = 0
            centered=centered,
            # to return gradR, gradZ, for D1N2 deg 0, for tomotok
            returnas_element=returnas_element,
        )

    def apply_bsplines_operator(
        self,
        key=None,
        operator=None,
        store=None,
        returnas=None,
    ):
        """ Apply an operator to desired data """

        raise NotImplementedError("TBD")
        return _operators.apply_operator(
            coll=self,
            key=key,
            operator=operator,
            store=store,
            returnas=returnas,
        )

    # -----------------
    # contours
    # ------------------

    def get_profile2d_contours(
        self,
        key=None,
        levels=None,
        res=None,
        npts=None,
        largest=None,
        ref_com=None,
        # return vs store
        returnas=None,
        return_dref=None,
        store=None,
        key_npts=None,
        key_cont0=None,
        key_cont1=None,
    ):
        """ Return a dict of contour data """

        return _contours.get_contours(
            coll=self,
            key=key,
            levels=levels,
            res=res,
            npts=npts,
            largest=largest,
            ref_com=ref_com,
            # return vs store
            returnas=returnas,
            return_dref=return_dref,
            store=store,
            key_npts=key_npts,
            key_cont0=key_cont0,
            key_cont1=key_cont1,
        )

    # -----------------
    # interp tools
    # ------------------

    def interpolate(
        self,
        # interpolation base
        keys=None,
        ref_key=None,
        # interpolation pts
        x0=None,
        x1=None,
        grid=None,
        res=None,
        mode=None,
        submesh=None,
        # domain limitation
        domain=None,
        # common ref
        ref_com=None,
        ref_vector_strategy=None,
        # bsplines-specific
        details=None,
        indbs_tf=None,
        # rect-specific
        crop=None,
        # parameters
        deg=None,
        deriv=None,
        val_out=None,
        log_log=None,
        nan0=None,
        # store vs return
        returnas=None,
        return_params=None,
        store=None,
        store_keys=None,
        inplace=None,
        # debug
        debug=None,
        # # bsplines
        # res=None,
        # imshow=None,
    ):
        """ Interpolate along a given ref or bspline, in 1d or 2d """

        return _interpolate.interpolate(
            coll=self,
            # interpolation base
            keys=keys,
            ref_key=ref_key,
            # interpolation pts
            x0=x0,
            x1=x1,
            grid=grid,
            res=res,
            mode=mode,
            submesh=submesh,
            # domain limitation
            domain=domain,
            # common ref
            ref_com=ref_com,
            ref_vector_strategy=ref_vector_strategy,
            # bsplines-specific
            details=details,
            indbs_tf=indbs_tf,
            # rect-specific
            crop=crop,
            # parameters
            deg=deg,
            deriv=deriv,
            val_out=val_out,
            log_log=log_log,
            # store vs return
            returnas=returnas,
            return_params=return_params,
            store=store,
            store_keys=store_keys,
            inplace=inplace,
            # debug
            debug=debug,
        )

    def interpolate_all_bsplines(
        self,
        keys=None,
        # sampling
        dres=None,
        dunique_mesh_2d=None,
        submesh=None,
        # parameters
        val_out=None,
        nan0=None,
    ):
        """ Return a new instance with fully interpolated data

        """

        return _interpolate_all.interpolate_all_bsplines(
            coll=self,
            keys=keys,
            # sampling
            dres=dres,
            dunique_mesh_2d=dunique_mesh_2d,
            submesh=submesh,
            # parameters
            val_out=val_out,
            nan0=nan0,
        )

    """
    def get_sample_bspline(self, key=None, res=None, grid=None, mode=None):
        return _compute.sample_bsplines(
            coll=self,
            key=key,
            res=res,
            grid=grid,
            mode=mode,
        )
    """

    # def _check_qr12RPZ(
        # self,
        # quant=None,
        # ref1d=None,
        # ref2d=None,
        # q2dR=None,
        # q2dPhi=None,
        # q2dZ=None,
        # group1d=None,
        # group2d=None,
    # ):

        # if group1d is None:
            # group1d = self._group1d
        # if group2d is None:
            # group2d = self._group2d

        # lc0 = [quant is None, ref1d is None, ref2d is None]
        # lc1 = [q2dR is None, q2dPhi is None, q2dZ is None]
        # if np.sum([all(lc0), all(lc1)]) != 1:
            # msg = (
                # "Please provide either (xor):\n"
                # + "\t- a scalar field (isotropic emissivity):\n"
                # + "\t\tquant : scalar quantity to interpolate\n"
                # + "\t\t\tif quant is 1d, intermediate reference\n"
                # + "\t\t\tfields are necessary for 2d interpolation\n"
                # + "\t\tref1d : 1d reference field on which to interpolate\n"
                # + "\t\tref2d : 2d reference field on which to interpolate\n"
                # + "\t- a vector (R,Phi,Z) field (anisotropic emissivity):\n"
                # + "\t\tq2dR :  R component of the vector field\n"
                # + "\t\tq2dPhi: R component of the vector field\n"
                # + "\t\tq2dZ :  Z component of the vector field\n"
                # + "\t\t=> all components have the same time and mesh!\n"
            # )
            # raise Exception(msg)

        # # Check requested quant is available in 2d or 1d
        # if all(lc1):
            # (
                # idquant, idref1d, idref2d,
            # ) = _compute._get_possible_ref12d(
                # dd=self._ddata,
                # key=quant, ref1d=ref1d, ref2d=ref2d,
                # group1d=group1d,
                # group2d=group2d,
            # )
            # idq2dR, idq2dPhi, idq2dZ = None, None, None
            # ani = False
        # else:
            # idq2dR, msg = _compute._get_keyingroup_ddata(
                # dd=self._ddata,
                # key=q2dR, group=group2d, msgstr='quant', raise_=True,
            # )
            # idq2dPhi, msg = _compute._get_keyingroup_ddata(
                # dd=self._ddata,
                # key=q2dPhi, group=group2d, msgstr='quant', raise_=True,
            # )
            # idq2dZ, msg = _compute._get_keyingroup_ddata(
                # dd=self._ddata,
                # key=q2dZ, group=group2d, msgstr='quant', raise_=True,
            # )
            # idquant, idref1d, idref2d = None, None, None
            # ani = True
        # return idquant, idref1d, idref2d, idq2dR, idq2dPhi, idq2dZ, ani

    # def _interp_pts2d_to_quant1d(
        # self,
        # pts=None,
        # vect=None,
        # t=None,
        # quant=None,
        # ref1d=None,
        # ref2d=None,
        # q2dR=None,
        # q2dPhi=None,
        # q2dZ=None,
        # interp_t=None,
        # interp_space=None,
        # fill_value=None,
        # Type=None,
        # group0d=None,
        # group1d=None,
        # group2d=None,
        # return_all=None,
    # ):
        # """ Return the value of the desired 1d quantity at 2d points

        # For the desired inputs points (pts):
            # - pts are in (X, Y, Z) coordinates
            # - space interpolation is linear on the 1d profiles
        # At the desired input times (t):
            # - using a nearest-neighbourg approach for time

        # """
        # # Check inputs
        # if group0d is None:
            # group0d = self._group0d
        # if group1d is None:
            # group1d = self._group1d
        # if group2d is None:
            # group2d = self._group2d
        # # msg = "Only 'nearest' available so far for interp_t!"
        # # assert interp_t == 'nearest', msg

        # # Check requested quant is available in 2d or 1d
        # idquant, idref1d, idref2d, idq2dR, idq2dPhi, idq2dZ, ani = \
                # self._check_qr12RPZ(
                    # quant=quant, ref1d=ref1d, ref2d=ref2d,
                    # q2dR=q2dR, q2dPhi=q2dPhi, q2dZ=q2dZ,
                    # group1d=group1d, group2d=group2d,
                # )

        # # Check the pts is (3,...) array of floats
        # idmesh = None
        # if pts is None:
            # # Identify mesh to get default points
            # if ani:
                # idmesh = [id_ for id_ in self._ddata[idq2dR]['ref']
                          # if self._dref[id_]['group'] == group2d][0]
            # else:
                # if idref1d is None:
                    # idmesh = [id_ for id_ in self._ddata[idquant]['ref']
                              # if self._dref[id_]['group'] == group2d][0]
                # else:
                    # idmesh = [id_ for id_ in self._ddata[idref2d]['ref']
                              # if self._dref[id_]['group'] == group2d][0]

            # # Derive pts
            # pts = self._get_pts_from_mesh(key=idmesh)

        # pts = np.atleast_2d(pts)
        # if pts.shape[0] != 3:
            # msg = (
                # "pts must be np.ndarray of (X,Y,Z) points coordinates\n"
                # + "Can be multi-dimensional, but 1st dimension is (X,Y,Z)\n"
                # + "    - Expected shape : (3,...)\n"
                # + "    - Provided shape : {}".format(pts.shape)
            # )
            # raise Exception(msg)

        # # Check t
        # lc = [t is None, type(t) is str, type(t) is np.ndarray]
        # assert any(lc)
        # if lc[1]:
            # assert t in self._ddata.keys()
            # t = self._ddata[t]['data']

        # # Interpolation (including time broadcasting)
        # # this is the second slowest step (~0.08 s)
        # func = self._get_finterp(
            # idquant=idquant, idref1d=idref1d, idref2d=idref2d,
            # idq2dR=idq2dR, idq2dPhi=idq2dPhi, idq2dZ=idq2dZ,
            # idmesh=idmesh,
            # interp_t=interp_t, interp_space=interp_space,
            # fill_value=fill_value, ani=ani, Type=Type,
            # group0d=group0d, group2d=group2d,
        # )

        # # Check vect of ani
        # c0 = (
            # ani is True
            # and (
                # vect is None
                # or not (
                    # isinstance(vect, np.ndarray)
                    # and vect.shape == pts.shape
                # )
            # )
        # )
        # if c0:
            # msg = (
                # "Anisotropic field interpolation needs a field of local vect\n"
                # + "  => Please provide vect as (3, npts) np.ndarray!"
            # )
            # raise Exception(msg)

        # # This is the slowest step (~1.8 s)
        # val, t = func(pts, vect=vect, t=t)

        # # return
        # if return_all is None:
            # return_all = True
        # if return_all is True:
            # dout = {
                # 't': t,
                # 'pts': pts,
                # 'ref1d': idref1d,
                # 'ref2d': idref2d,
                # 'q2dR': idq2dR,
                # 'q2dPhi': idq2dPhi,
                # 'q2dZ': idq2dZ,
                # 'interp_t': interp_t,
                # 'interp_space': interp_space,
            # }
            # return val, dout
        # else:
            # return val

    # def _interp_pts2d_to_quant1d(
        # self,
        # pts=None,
        # vect=None,
        # t=None,
        # quant=None,
        # ref1d=None,
        # ref2d=None,
        # q2dR=None,
        # q2dPhi=None,
        # q2dZ=None,
        # interp_t=None,
        # interp_space=None,
        # fill_value=None,
        # Type=None,
        # group0d=None,
        # group1d=None,
        # group2d=None,
        # return_all=None,
    # ):
        # """ Return the value of the desired 1d quantity at 2d points

        # For the desired inputs points (pts):
            # - pts are in (X, Y, Z) coordinates
            # - space interpolation is linear on the 1d profiles
        # At the desired input times (t):
            # - using a nearest-neighbourg approach for time

        # """
        # # Check inputs
        # if group0d is None:
            # group0d = self._group0d
        # if group1d is None:
            # group1d = self._group1d
        # if group2d is None:
            # group2d = self._group2d
        # # msg = "Only 'nearest' available so far for interp_t!"
        # # assert interp_t == 'nearest', msg

        # # Check requested quant is available in 2d or 1d
        # idquant, idref1d, idref2d, idq2dR, idq2dPhi, idq2dZ, ani = \
                # self._check_qr12RPZ(
                    # quant=quant, ref1d=ref1d, ref2d=ref2d,
                    # q2dR=q2dR, q2dPhi=q2dPhi, q2dZ=q2dZ,
                    # group1d=group1d, group2d=group2d,
                # )

        # # Check the pts is (3,...) array of floats
        # idmesh = None
        # if pts is None:
            # # Identify mesh to get default points
            # if ani:
                # idmesh = [id_ for id_ in self._ddata[idq2dR]['ref']
                          # if self._dref[id_]['group'] == group2d][0]
            # else:
                # if idref1d is None:
                    # idmesh = [id_ for id_ in self._ddata[idquant]['ref']
                              # if self._dref[id_]['group'] == group2d][0]
                # else:
                    # idmesh = [id_ for id_ in self._ddata[idref2d]['ref']
                              # if self._dref[id_]['group'] == group2d][0]

            # # Derive pts
            # pts = self._get_pts_from_mesh(key=idmesh)

        # pts = np.atleast_2d(pts)
        # if pts.shape[0] != 3:
            # msg = (
                # "pts must be np.ndarray of (X,Y,Z) points coordinates\n"
                # + "Can be multi-dimensional, but 1st dimension is (X,Y,Z)\n"
                # + "    - Expected shape : (3,...)\n"
                # + "    - Provided shape : {}".format(pts.shape)
            # )
            # raise Exception(msg)

        # # Check t
        # lc = [t is None, type(t) is str, type(t) is np.ndarray]
        # assert any(lc)
        # if lc[1]:
            # assert t in self._ddata.keys()
            # t = self._ddata[t]['data']

        # # Interpolation (including time broadcasting)
        # # this is the second slowest step (~0.08 s)
        # func = self._get_finterp(
            # idquant=idquant, idref1d=idref1d, idref2d=idref2d,
            # idq2dR=idq2dR, idq2dPhi=idq2dPhi, idq2dZ=idq2dZ,
            # idmesh=idmesh,
            # interp_t=interp_t, interp_space=interp_space,
            # fill_value=fill_value, ani=ani, Type=Type,
            # group0d=group0d, group2d=group2d,
        # )

        # # Check vect of ani
        # c0 = (
            # ani is True
            # and (
                # vect is None
                # or not (
                    # isinstance(vect, np.ndarray)
                    # and vect.shape == pts.shape
                # )
            # )
        # )
        # if c0:
            # msg = (
                # "Anisotropic field interpolation needs a field of local vect\n"
                # + "  => Please provide vect as (3, npts) np.ndarray!"
            # )
            # raise Exception(msg)

        # # This is the slowest step (~1.8 s)
        # val, t = func(pts, vect=vect, t=t)

        # # return
        # if return_all is None:
            # return_all = True
        # if return_all is True:
            # dout = {
                # 't': t,
                # 'pts': pts,
                # 'ref1d': idref1d,
                # 'ref2d': idref2d,
                # 'q2dR': idq2dR,
                # 'q2dPhi': idq2dPhi,
                # 'q2dZ': idq2dZ,
                # 'interp_t': interp_t,
                # 'interp_space': interp_space,
            # }
            # return val, dout
        # else:
            # return val

    # def interpolate_profile2d(
        # # ressources
        # self,
        # # interpolation base, 1d or 2d
        # key=None,
        # # external coefs (optional)
        # coefs=None,
        # # interpolation points
        # R=None,
        # Z=None,
        # radius=None,
        # angle=None,
        # grid=None,
        # radius_vs_time=None,
        # azone=None,
        # # time: t or indt
        # t=None,
        # indt=None,
        # indt_strict=None,
        # # bsplines
        # indbs=None,
        # # parameters
        # details=None,
        # reshape=None,
        # res=None,
        # crop=None,
        # nan0=None,
        # val_out=None,
        # imshow=None,
        # return_params=None,
        # # storing
        # store=None,
        # inplace=None,
    # ):
        # """ Interpolate desired profile2d (i.e.: data on bsplines)

        # Interpolate:
            # - key: a data on bsplines
            # - coefs: external-provided set of coefs

        # coefs can only be provided if:
            # - details = False
            # - key = keybs
            # - coefs is a scalar or has shape = shapebs

        # At points:
            # - R:  R coordinates (np.ndarray or scalar)
            # - Z:  Z coordinates (np.ndarray, same shape as R, or scalar)
            # - grid: bool, if True mesh R x Z
            # - indt: if provided, only interpolate at desired time indices

        # With options:
            # - details: bool, if True returns value for each bspline
            # - indbs:   optional, select bsplines for which to interpolate
            # - reshape: bool,
            # - res:  optional, resolution to generate R and Z if they are None
            # - crop: bool, whether to use the cropped mesh
            # - nan0: value for out-of-mesh points
            # - imshow: bool, whether to return as imshow (transpose)
            # - return_params: bool, whether to return dict of input params

        # """

        # return _compute.interp2d(
            # # ressources
            # coll=self,
            # # interpolation base, 1d or 2d
            # key=key,
            # # external coefs (optional)
            # coefs=coefs,
            # # interpolation points
            # R=R,
            # Z=Z,
            # radius=radius,
            # angle=angle,
            # grid=grid,
            # radius_vs_time=radius_vs_time,
            # azone=azone,
            # # time: t or indt
            # t=t,
            # indt=indt,
            # indt_strict=indt_strict,
            # # bsplines
            # indbs=indbs,
            # # parameters
            # details=details,
            # reshape=reshape,
            # res=res,
            # crop=crop,
            # nan0=nan0,
            # val_out=val_out,
            # imshow=imshow,
            # return_params=return_params,
            # # storing
            # store=store,
            # inplace=inplace,
        # )

    # TBF after polar meshes
    # def interpolate_2dto1d(
        # resources
        # self,
        # interpolation base
        # key1d=None,
        # key2d=None,
        # interpolation pts
        # R=None,
        # Z=None,
        # grid=None,
        # parameters
        # interp_t=None,
        # fill_value=None,
        # ani=False,
    # ):

        # return _compute.interp2dto1d(
            # coll=self,
            # key1d=key1d,
            # key2d=key2d,
            # R=R,
            # Z=Z,
            # grid=grid,
            # crop=crop,
            # nan0=nan0,
            # return_params=return_params,
        # )

    # -----------------
    # plotting
    # ------------------

    def plot_as_profile2d(
        self,
        key=None,
        # parameters
        dres=None,
        dunique_mesh_2d=None,
        # levels
        dlevels=None,
        ref_com=None,
        # details
        plot_details=None,
        # ref vectors
        dref_vectorZ=None,
        dref_vectorU=None,
        ref_vector_strategy=None,
        # interpolation
        val_out=None,
        nan0=None,
        # plot options
        vmin=None,
        vmax=None,
        cmap=None,
        dax=None,
        dmargin=None,
        fs=None,
        dcolorbar=None,
        dleg=None,
        # interactivity
        dinc=None,
        connect=None,
    ):

        return _plot_as_profile2d.plot_as_profile2d(
            coll=self,
            key=key,
            # parameters
            dres=dres,
            dunique_mesh_2d=dunique_mesh_2d,
            # levels
            dlevels=dlevels,
            ref_com=ref_com,
            # details
            plot_details=plot_details,
            # ref vectors
            dref_vectorZ=dref_vectorZ,
            dref_vectorU=dref_vectorU,
            ref_vector_strategy=ref_vector_strategy,
            # interpolation
            val_out=val_out,
            nan0=nan0,
            # plot options
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            dcolorbar=dcolorbar,
            dleg=dleg,
            # interactivity
            dinc=dinc,
            connect=connect,
        )

    def plot_as_profile2d_compare(
        self,
        keys=None,
        # parameters
        dres=None,
        # levels
        dlevels=None,
        ref_com=None,
        # details
        plot_details=None,
        # ref vectors
        dref_vectorZ=None,
        dref_vectorU=None,
        # interpolation
        val_out=None,
        nan0=None,
        # plot options
        vmin=None,
        vmax=None,
        cmap=None,
        dax=None,
        dmargin=None,
        fs=None,
        dcolorbar=None,
        dleg=None,
        # interactivity
        dinc=None,
        connect=None,
    ):

        return _plot_as_profile2d_compare.plot_as_profile2d_compare(
            coll=self,
            keys=keys,
            # parameters
            dres=dres,
            # levels
            dlevels=dlevels,
            ref_com=ref_com,
            # details
            plot_details=plot_details,
            # ref vectors
            dref_vectorZ=dref_vectorZ,
            dref_vectorU=dref_vectorU,
            # interpolation
            val_out=val_out,
            nan0=nan0,
            # plot options
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            dcolorbar=dcolorbar,
            dleg=dleg,
            # interactivity
            dinc=dinc,
            connect=connect,
        )


    # def plot_bsplines(
        # self,
        # key=None,
        # indbs=None,
        # indt=None,
        # knots=None,
        # cents=None,
        # res=None,
        # plot_mesh=None,
        # val_out=None,
        # nan0=None,
        # cmap=None,
        # dax=None,
        # dmargin=None,
        # fs=None,
        # dleg=None,
    # ):

        # return _plot.plot_bspline(
            # coll=self,
            # key=key,
            # indbs=indbs,
            # indt=indt,
            # knots=knots,
            # cents=cents,
            # res=res,
            # plot_mesh=plot_mesh,
            # val_out=val_out,
            # nan0=nan0,
            # cmap=cmap,
            # dax=dax,
            # dmargin=dmargin,
            # fs=fs,
            # dleg=dleg,
        # )



""" A tomography library for fusion devices

See:
https://github.com/ToFuProject/bsplines2d
"""

# Built-in
import os
import subprocess
from codecs import open
# ... setup tools
from setuptools import setup, find_packages


# ... local script
import _updateversion as up


# == Getting version =====================================================
_HERE = os.path.abspath(os.path.dirname(__file__))

version = up.updateversion()

print("")
print("Version for setup.py : ", version)
print("")


# =============================================================================
# Get the long description from the README file
# Get the readme file whatever its extension (md vs rst)

_README = [
    ff
    for ff in os.listdir(_HERE)
    if len(ff) <= 10 and ff[:7] == "README."
]
assert len(_README) == 1
_README = _README[0]
with open(os.path.join(_HERE, _README), encoding="utf-8") as f:
    long_description = f.read()
if _README.endswith(".md"):
    long_description_content_type = "text/markdown"
else:
    long_description_content_type = "text/x-rst"


# =============================================================================


# =============================================================================
#  Compiling files

setup(
    name="bsplines2d",
    version=f"{version}",
    # Use scm to get code version from git tags
    # cf. https://pypi.python.org/pypi/setuptools_scm
    # Versions should comply with PEP440. For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    # The version is stored only in the setup.py file and read from it (option
    # 1 in https://packaging.python.org/en/latest/single_source_version.html)
    use_scm_version=False,

    # Description of what library does
    description="A python library for generic multidimensional bsplines on various meshes, using datastock",
    long_description=long_description,
    long_description_content_type=long_description_content_type,

    # The project's main homepage.
    url="https://github.com/ToFuProject/bsplines2d",
    # Author details
    author="Didier VEZINET",
    author_email="didier.vezinet@gmail.com",

    # Choose your license
    license="MIT",

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        # 3 - Alpha
        # 4 - Beta
        # 5 - Production/Stable
        "Development Status :: 4 - Beta",
        # Indicate who your project is intended for
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Physics",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        # In which language most of the code is written ?
        "Natural Language :: English",
    ],

    # What does your project relate to?
    keywords="bsplines data analysis modelling mesh plot",

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(
        exclude=[
            "doc",
        ]
    ),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    # py_modules=["my_module"],
    # List run-time dependencies here. These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        "numpy",
        "scipy",
        "matplotlib",
        "datastock>=0.0.30",
    ],
    python_requires=">=3.6",

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        "dev": [
            "check-manifest",
            "coverage",
            "pytest",
            "sphinx",
            "sphinx-gallery",
            "sphinx_bootstrap_theme",
        ]
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here. If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    # package_data={
    #    # If any package contains *.txt, *.rst or *.npz files, include them:
    #    '': ['*.txt', '*.rst', '*.npz'],
    #    # And include any *.csv files found in the 'ITER' package, too:
    #    'ITER': ['*.csv'],
    # },
    package_data={
        'bsplines2d.tests.test_data': ['.npz'],
    },
    include_package_data=True,

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html
    # installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],

    # executable scripts can be declared here
    # They can be python or non-python scripts
    # scripts=[
    # ],

    # entry_points point to functions in the package
    # Theye are generally preferable over scripts because they provide
    # cross-platform support and allow pip to create the appropriate form
    # of executable for the target platform.
    # entry_points={},
    # include_dirs=[np.get_include()],

    py_modules=['_updateversion'],
)



# -*- coding: utf-8 -*-


# Built-in
import itertools as itt


import numpy as np
import scipy.interpolate as scpinterp


from . import _utils_bsplines
from . import _class02_bsplines_operators_1d as _operators_1d


# ################################################################
# ################################################################
#                   class
# ################################################################


class UnivariateSpline():
    """ Subclass for tofu

    Defined from knots (unique) and deg
    coefs set to 1 by default

    Used self.set_coefs() to update
    """

    def __init__(self, knots=None, deg=None):

        assert deg in [0, 1, 2, 3], deg

        # get knots pr bs
        self._get_knots_per_bs_for_basis_elements(
            knots=knots,
            deg=deg,
        )

        # get nbs
        self.shapebs = (self.nbs,)

        # deg
        self.deg = deg

    def _get_knots_per_bs_for_basis_elements(
        self,
        knots=None,
        deg=None,
    ):

        # ------------------------------------
        # get knots per bs in radius direction

        knots_per_bs = _utils_bsplines._get_knots_per_bs(
            knots, deg=deg, returnas='data',
        )
        knots_with_mult, nbs = _utils_bsplines._get_knots_per_bs(
            knots, deg=deg, returnas='data', return_unique=True,
        )

        if nbs != knots_per_bs.shape[1]:
            msg = "Inconsistent nb. of splines"
            raise Exception(msg)

        # ----------------
        # Pre-compute bsplines basis elements

        lbs = [
            scpinterp.BSpline.basis_element(
                knots_per_bs[:, ii],
                extrapolate=False,
            )
            for ii in range(nbs)
        ]

        # ----------------
        # bsplines centers

        cents_per_bs = _utils_bsplines._get_cents_per_bs(
            0.5*(knots[1:] + knots[:-1]),
            deg=deg,
            returnas='data',
        )

        # ----------------
        # bsplines apex

        apex_per_bs = _utils_bsplines._get_apex_per_bs(
            knots=knots,
            knots_per_bs=knots_per_bs,
            deg=deg,
        )

        # ------
        # store

        self.knots = knots
        self.knots_with_mult = knots_with_mult
        self.knots_per_bs = knots_per_bs
        self.cents_per_bs = cents_per_bs
        self.apex_per_bs = apex_per_bs
        self.nbs = nbs
        self.lbs = lbs

    def _check_coefs(self, coefs=None, axis=None):
        """ None for ev_details, (nt, shapebs) for sum """
        if coefs.shape[axis[0]] != self.nbs:
            msg = (
                "Arg coefs has wrong shape!\n"
                f"\t- coefs.shape = {coefs.shape}\n"
                f"\t- coefs.shape[{axis[0]}] != {self.nbs}\n"
            )
            raise Exception(msg)

    def __call__(
        self,
        # interp points
        x0=None,
        # coefs
        coefs=None,
        axis=None,
        # options
        val_out=None,
        deriv=None,
        # slicing
        sli_c=None,
        sli_x=None,
        sli_v=None,
        sli_o=None,
        indokx0=None,
        shape_v=None,
        shape_o=None,
        dref_com=None,
        # for compatibility (unused)
        **kwdargs,
    ):
        """ Assumes

        coefs.shape = (..., nbs, ...)

        """

        # ------------
        # check inputs

        # coefs
        self._check_coefs(coefs=coefs, axis=axis)

        # x0, shape, val
        val = np.zeros(shape_v, dtype=float)

        # ------------
        # compute

        for ind in itt.product(*[range(aa) for aa in shape_o]):

            # slices
            slic = sli_c(
                ind,
                axis=axis,
                ddim=coefs.ndim,
            )

            slix = sli_x(
                ind,
                indokx0=indokx0,
                ix=dref_com['ix'],
                iother=dref_com['iother'],
            )

            sliv = sli_v(
                ind,
                indokx0=indokx0,
                ddim=coefs.ndim,
                axis=axis,
                ix=dref_com['ix'],
                iother=dref_com['iother'],
            )

            # call be called on any shape of x0
            val[sliv] = scpinterp.BSpline(
                self.knots_with_mult,
                coefs[slic],
                self.deg,
                axis=0,
                extrapolate=False,
            )(x0[slix], nu=deriv)

        # clean out-of-mesh
        if dref_com['ix'] is None and val_out is not False:
            slio = sli_o((x0 < self.knots[0]) | (x0 > self.knots[-1]))
            val[slio] = val_out

        return val

    # TBF
    def ev_details(
        self,
        # coordinates
        x0=None,
        # options
        indbs_tf=None,
        deriv=None,
        val_out=None,
        # for compatibility (unused)
        **kwdargs,
    ):
        """ Assumes

        indbs_tf = flat array of int indices

        """

        # ------------
        # check inputs

        if indbs_tf is None:
            nbs = self.nbs
        else:
            indbs_tf = np.atleast_1d(indbs_tf).ravel()
            assert 'int' in indbs_tf.dtype.name, indbs_tf
            assert np.unique(indbs_tf).size == indbs_tf.size
            nbs = indbs_tf.size

        # -------
        # prepare

        shape = tuple(np.r_[x0.shape, nbs])
        val = np.full(shape, val_out)

        # ------------
        # compute

        ni = 0
        for ii in range(self.nbs):

            if indbs_tf is not None and ii not in indbs_tf:
                continue

            iok = (
                (x0 >= self.knots_per_bs[0, ii])
                & ((x0 < self.knots_per_bs[-1, ii]))
            )
            if np.any(iok):
                val[iok, ni] = self.lbs[ii](x0[iok], nu=deriv)
            ni += 1

        return val

    # -----------------
    # constraints methods
    # -----------------

    def get_constraints_out_rlim(
        self,
        rlim=None,
        rm=None,
        lim=None,
    ):
        """
        Return indices of bslines fully out of rlim (min or max)

        Assumes rlim is a 1d array, one per time step
        """


        # ------------
        # check inputs

        if rm not in ['rmin', 'rmax']:
            msg = f"Invalid rm!\nShould be in ['rmax', 'rmin']\nProvided: {rm}"
            raise Exception(msg)

        if np.isscalar(rlim):
            rlim = np.r_[rlim]

        if not (isinstance(rlim, np.ndarray) and rlim.ndim == 1):
            msg = "rlim must ba a 1d array!"
            raise Exception(msg)

        # ------------
        # get index of bsplines out of limits

        kpbs = self.knots_per_bs
        if rm == 'rmax':
            if lim == 'allout':
                ind = np.all(
                    kpbs[None, :, :] > rlim[:, None, None],
                    axis=1,
                )
            elif lim == 'outer':
                ind = np.sum(
                    kpbs[None, :, :] > rlim[:, None, None],
                    axis=1,
                ) > 1
            elif lim == 'inner':
                ind = np.any(
                    kpbs[None, :, :] > rlim[:, None, None],
                    axis=1,
                )

        else:
            if lim == 'allin':
                ind = np.all(
                    kpbs[None, :, :] < rlim[:, None, None],
                    axis=1,
                )
            elif lim == 'inner':
                ind = np.sum(
                    kpbs[None, :, :] < rlim[:, None, None],
                    axis=1,
                ) > 1
            elif lim == 'outer':
                ind = np.any(
                    kpbs[None, :, :] < rlim[:, None, None],
                    axis=1,
                )

        # if self.knotsa is None:
            # pass
        # else:
            # ind = np.repeat(ind, self.nbs_a_per_r, axis=1)

        # ------------
        # get coefs / offset

        offset = np.zeros((self.nbs,), dtype=float)

        return ind, offset

    def get_constraints_deriv(
        self,
        deriv=None,
        x0=None,
        val=None,
    ):
        """
        To set constraints on a derivative
        Return indices of bsplines + coefs + offset

        Assumes:
            - deriv in ['deriv0', 'deriv1']
            - rad and val are 1d arrays of the same shape

        return as flattened nbsplines indexing

        """
        # ------------
        # check inputs

        ld = ['deriv0', 'deriv1']
        if deriv not in ld:
            msg = f"Arg deriv must be in {ld}!\n Provided: {deriv}"
            raise Exception(msg)

        # --------
        # compute

        # coefs per radius per bs (nrad, nbs)
        ideriv = int(deriv[-1])
        vv = self.ev_details(x0=x0, deriv=ideriv, val_out=0.)

        # check conflicts
        indok = (vv != 0)
        if np.unique(indok, axis=0).shape[0] < indok.shape[0]:
            msg = f"Conflicting constraints on {deriv}:\n{indok}"
            raise Exception(msg)

        coefs = vv
        offset = np.repeat(val[:, None], self.nbs, axis=1)

        return indok, coefs, offset

    # -----------------
    # operator methods
    # -----------------

    def get_overlap(self):
        return _get_overlap(
            deg=self.deg,
            knots=self.knots_per_bs,
            shapebs=self.shapebs,
        )

    def get_operator(
        self,
        operator=None,
        geometry=None,
        cropbs_flat=None,
        # specific to deg = 0
        cropbs=None,
        centered=None,
        # to return gradR, gradZ, for D1N2 deg 0, for tomotok
        returnas_element=None,
    ):
        """ Get desired operator """

        return _operators_1d.get_operators(
            deg=self.deg,
            operator=operator,
            geometry=geometry,
            knots_mult=self.knots_with_mult,
            knots_per_bs=self.knots_per_bs,
            overlap=self.get_overlap(),
            # specific to deg = 0
            centered=centered,
            # to return gradR, gradZ, for D1N2 deg 0, for tomotok
            returnas_element=returnas_element,
        )


# ################################################################
# ################################################################
#                       overlap
# ################################################################


def _get_overlap(
    deg=None,
    knots=None,
    shapebs=None,
):
    # nb of overlapping, inc. itself in 1d
    nbs, = shapebs
    ind0 = np.arange(0, nbs)

    # complete
    ntot = 2*deg + 1

    add0 = np.arange(-deg, deg+1)

    inter = ind0[None, :] + add0[:, None]

    # purge
    ineg = (inter < 0) | (inter >= nbs)
    inter[ineg] = -1

    return inter


# ################################################################
# ################################################################
#                   Main
# ################################################################


def get_bs_class(
    deg=None,
    knots=None,
):

    # ----------------
    # Define functions

    return UnivariateSpline(
        knots=knots,
        deg=deg,
    )



# -*- coding: utf-8 -*-


# Built-in
import warnings


# Common
import numpy as np
from matplotlib.tri import Triangulation as mplTri
import scipy.spatial as scpsp
import datastock as ds


# specific
from . import _generic_mesh


# ################################################################
# ################################################################
#                           mesh generic check
# ################################################################


def check(
    coll=None,
    key=None,
    # knots
    knots=None,
    indices=None,
    # from pts
    pts_x0=None,
    pts_x1=None,
    # names of coords
    knots0_name=None,
    knots1_name=None,
    # attributes
    **kwdargs,
):

    # key
    key = ds._generic_check._obj_key(
        d0=coll._dobj.get(coll._which_mesh, {}),
        short='m',
        key=key,
    )

    # -------------
    # knots vs pts

    lc = [
        knots is not None and indices is not None,
        pts_x0 is not None and pts_x1 is not None,
    ]
    if np.sum(lc) != 1:
        msg = (
            "Please provide args (knots, indices) xor (pts_x0, pts_x1)!\n"
            f"\t- knots: {knots}\n"
            f"\t- indices: {indices}\n"
            f"\t- pts_x0: {pts_x0}\n"
            f"\t- pts_x1: {pts_x1}\n"
        )
        raise Exception(msg)

    # --------------------
    # from pts and polygon

    if lc[1]:
        knots, indices = _from_pts_poly(
            pts_x0=pts_x0,
            pts_x1=pts_x1,
        )

    # -------------
    # knots vectors

    knots, indices, ntri = _check_knotscents(
        key=key,
        knots=knots,
        indices=indices,
    )

    cents0 = np.mean(knots[indices, 0], axis=1)
    cents1 = np.mean(knots[indices, 1], axis=1)

    # --------------
    # to dict

    dref, ddata, dobj = _to_dict(
        coll=coll,
        key=key,
        knots=knots,
        indices=indices,
        cents0=cents0,
        cents1=cents1,
        ntri=ntri,
        knots0_name=knots0_name,
        knots1_name=knots1_name,
        **kwdargs,
    )

    return key, dref, ddata, dobj


# ################################################################
# ################################################################
#                      from pts and polygon
# ################################################################


def _from_pts_poly(
    pts_x0=None,
    pts_x1=None,
):

    pts_x0 = ds._generic_check._check_flat1darray(
        pts_x0, 'pts_x0',
        dtype=float,
        can_be_None=False,
    )

    pts_x1 = ds._generic_check._check_flat1darray(
        pts_x1, 'pts_x1',
        dtype=float,
        size=pts_x0.size,
        can_be_None=False,
    )

    # --------
    # Delauney

    knots = np.array([pts_x0, pts_x1]).T

    delau = scpsp.Delaunay(
        knots,
        furthest_site=False,
        incremental=False,
        qhull_options='QJ',
    )

    return knots, delau.simplices


# ################################################################
# ################################################################
#                          check knots and indices
# ################################################################


def _check_knotscents(
    key=None,
    knots=None,
    indices=None,
):

    # ---------------------
    # check mesh conformity

    indices, knots = _mesh2DTri_conformity(knots=knots, indices=indices, key=key)

    # ---------------------------------------------
    # define triangular mesh and trifinder function

    # triangular mesh
    if indices.shape[1] == 3:

        # check clock-wise triangles
        indices = _mesh2DTri_ccw(knots=knots, indices=indices, key=key)
        ntri = 1

    # Quadrangular mesh => convert to triangular
    elif indices.shape[1] == 4:

        ind2 = np.empty((indices.shape[0]*2, 3), dtype=int)
        ind2[::2, :] = indices[:, :3]
        ind2[1::2, :-1] = indices[:, 2:]
        ind2[1::2, -1] = indices[:, 0]
        indices = ind2

        # Re-check mesh conformity
        indices, knots = _mesh2DTri_conformity(knots=knots, indices=indices, key=key)
        indices = _mesh2DTri_ccw(knots=knots, indices=indices, key=key)
        ntri = 2

    # ----------------------------
    # Check on trifinder function

    out = 'not done'
    try:
        trifind = mplTri(knots[:, 0], knots[:, 1], indices).get_trifinder()
        out = trifind(np.r_[0.], np.r_[0])
        assert isinstance(out, np.ndarray)
    except Exception as err:
        msg = str(err) + (
            "\nArg trifind must return an array of indices when fed with arrays "
            "of (R, Z) coordinates!\n"
            f"\ttrifind(np.r_[0], np.r_[0.]) = {out}\n"
            f"\t- ntri = {ntri}\n"
            f"\t- indices = {indices}\n"
            f"\t- knots = {knots}"
        )
        raise Exception(msg)

    return knots, indices, ntri


# ################################################################
# ################################################################
#                           Mesh2DTri
# ################################################################


def _mesh2DTri_conformity(knots=None, indices=None, key=None):

    # ---------------------------------
    # make sure np.ndarrays of dim = 2

    knots = np.atleast_2d(knots).astype(float)
    indices = np.atleast_2d(indices).astype(int)

    # --------------
    # check shapes

    c0 = (
        knots.shape[1] == 2
        and knots.shape[0] >= 3
        and indices.shape[1] in [3, 4]
        and indices.shape[0] >= 1
        and indices.dtype == int
    )
    if not c0:
        msg = (
            "Arg knots must be of shape (nknots>=3, 2) and "
            "arg indices must be of shape (nind>=1, 3 or 4) and dtype = int\n"
            "Provided:\n"
            f"\t- knots.shape: {knots.shape}\n"
            f"\t- indices.shape: {indices.shape}\n"
            f"\t- indices.dtype: {indices.dtype}\n"
        )
        raise Exception(msg)

    nknots = knots.shape[0]
    nind = indices.shape[0]

    # -------------------
    # Test for duplicates

    # knots (floats => distance)
    dist = np.full((nknots, nknots), np.nan)
    ind = np.zeros(dist.shape, dtype=bool)
    for ii in range(nknots):
        dist[ii, ii+1:] = np.sqrt(
            (knots[ii+1:, 0] - knots[ii, 0])**2
            + (knots[ii+1:, 1] - knots[ii, 1])**2
        )
        ind[ii, ii+1:] = True

    ind[ind] = dist[ind] < 1.e-6
    if np.any(ind):
        iind = np.any(ind, axis=1).nonzero()[0]
        lstr = [f'\t\t- {ii}: {ind[ii, :].nonzero()[0]}' for ii in iind]
        msg = (
            f"Non-valid mesh {key}: \n"
            f"  Duplicate knots: {ind.sum()}\n"
            f"\t- knots.shape: {indices.shape}\n"
            f"\t- duplicate indices:\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    # cents (indices)
    indu = np.unique(indices, axis=0)
    if indu.shape[0] != nind:
        msg = (
            f"Non-valid mesh {key}: \n"
            f"  Duplicate indices: {nind - indu.shape[0]}\n"
            f"\t- indices.shape: {indices.shape}\n"
            f"\t- unique shape: {indu.shape}"
        )
        raise Exception(msg)

    # -------------------------------
    # Test for unused / unknown knots

    indu = np.unique(indu)
    c0 = np.all(indu >= 0) and indu.size == nknots

    # unused knots
    ino = (~np.in1d(
        range(0, nknots),
        indu,
        assume_unique=False,
        invert=False,
    )).nonzero()[0]

    # unknown knots
    unknown = np.setdiff1d(indu, range(nknots), assume_unique=True)

    if ino.size > 0 or unknown.size > 0:
        msg = f"Knots non-conformity identified for triangular mesh '{key}':\n"
        if ino.size > 0:
            msg += f"\t- Unused knots indices: {ino}\n"
        if unknown.size > 0:
            msg += f"\t- Unknown knots indices: {unknown}\n"
        raise Exception(msg)

    if indu.size < nknots:
        msg = (
            f"Unused knots in {key}:\n"
            f"\t- unused knots indices: {ino}"
        )
        warnings.warn(msg)

    elif indu.size > nknots or indu.max() != nknots - 1:
        unknown = np.setdiff1d(indu, range(nknots), assume_unique=True)
        msg = (
            "Unknown knots refered to in indices!\n"
            f"\t- unknown knots: {unknown}"
        )
        raise Exception(msg)

    return indices, knots


def _mesh2DTri_ccw(knots=None, indices=None, key=None):

    x, y = knots[indices, 0], knots[indices, 1]
    orient = (
        (y[:, 1] - y[:, 0])*(x[:, 2] - x[:, 1])
        - (y[:, 2] - y[:, 1])*(x[:, 1] - x[:, 0])
    )

    clock = orient > 0.
    if np.any(clock):
        msg = (
            "Some triangles are not counter-clockwise\n"
            "  (necessary for matplotlib.tri.Triangulation)\n"
            f"\t=> {clock.sum()} / {indices.shape[0]} triangles reshaped\n"
            f"\t indices: {clock.nonzero()[0]}"
        )
        warnings.warn(msg)
        indices[clock, 1], indices[clock, 2] = indices[clock, 2], indices[clock, 1]
    return indices


def _to_dict(
    coll=None,
    key=None,
    knots=None,
    indices=None,
    cents0=None,
    cents1=None,
    ntri=None,
    # names of coords
    knots0_name=None,
    knots1_name=None,
    # attributes
    **kwdargs,
):

    # ---------
    # check

    knots0_name = ds._generic_check._check_var(
        knots0_name, 'knots0_name',
        types=str,
        default='x',
    )

    knots1_name = ds._generic_check._check_var(
        knots1_name, 'knots1_name',
        types=str,
        default='y',
        excluded=[knots0_name],
    )

    # -----------------
    # keys

    kk = f"{key}-nk"
    kc = f"{key}-nc"
    ki = f"{key}-nind"

    _, _, kk0, kc0 = _generic_mesh.names_knots_cents(
        key=key,
        knots_name='0',
    )
    _, _, kk1, kc1 = _generic_mesh.names_knots_cents(
        key=key,
        knots_name='1',
    )

    kii = f"{key}-ind"

    # attributes
    latt = ['dim', 'quant', 'name', 'units']
    dim, quant, name, units = _generic_mesh._get_kwdargs_2d(kwdargs, latt)

    # -----------------
    # dict

    # dref
    dref = {
        kk: {
            'size': knots.shape[0],
        },
        kc: {
            'size': indices.shape[0],
        },
        ki: {
            'size': 3,
        },
    }

    # ddata
    ddata = {
        kk0: {
            'data': knots[:, 0],
            'units': units[0],
            'quant': quant[0],
            'dim': dim[0],
            'name': name[0],
            'ref': kk,
        },
        kk1: {
            'data': knots[:, 1],
            'units': units[1],
            'quant': quant[1],
            'dim': dim[1],
            'name': name[1],
            'ref': kk,
        },
        kc0: {
            'data': cents0,
            'units': units[0],
            'quant': quant[0],
            'dim': dim[0],
            'name': name[0],
            'ref': kc,
        },
        kc1: {
            'data': cents1,
            'units': units[1],
            'quant': quant[1],
            'dim': dim[1],
            'name': name[1],
            'ref': kc,
        },
        kii: {
            'data': indices,
            # 'units': '',
            'quant': 'indices',
            'dim': 'indices',
            'ref': (kc, ki),
        },
    }

    # dobj
    dobj = {
        coll._which_mesh: {
            key: {
                'nd': '2d',
                'type': 'tri',
                'ntri': ntri,
                'cents': (kc0, kc1),
                'knots': (kk0, kk1),
                'ind': kii,
                # 'ref-k': (kk,),
                # 'ref-c': (kc,),
                'shape-c': (indices.shape[0],),
                'shape-k': (knots.shape[0],),
                'crop': False,
            },
        }
    }

    # additional attributes
    for k0, v0 in kwdargs.items():
        if k0 not in latt:
            dobj[coll._which_mesh][key][k0] = v0

    return dref, ddata, dobj



# -*- coding: utf-8 -*-


# Common
import datastock as ds


# local
from . import _generic_mesh
from . import _class01_checks_1d as _checks_1d
from . import _class01_checks_2d_rect as _checks_2d_rect
from . import _class01_checks_2d_tri as _checks_2d_tri
# from . import _class01_checks_2d_polar as _checks_2d_polar


from . import _class01_rect_cropping as _cropping
from . import _class01_select as _select
from . import _class01_sample as _sample
from . import _class01_outline as _outline
from . import _class01_plot as _plot


__all__ = ['Mesh2D']


# #############################################################################
# #############################################################################
#
# #############################################################################


class Mesh2D(ds.DataStock):

    _which_mesh = 'mesh'
    _which_bsplines = 'bsplines'

    _show_in_summary = 'all'
    _dshow = dict(ds.DataStock._dshow)
    _dshow.update({
        _which_mesh: [
            'nd',
            'type',
            'variable',
            'crop',
            'crop-thresh',
            'ntri',
            'knots',
            'cents',
            'shape-k',
            'shape-c',
            'ind',
            'subkey',
            f'nb {_which_bsplines}',
        ],
    })

    def add_mesh_1d(
        self,
        key=None,
        knots=None,
        uniform=None,
        # defined from pre-existing bsplines
        subkey=None,
        # direct addition of bsplines
        deg=None,
        # additional attributes
        **kwdargs,
    ):

        """ Add an 1d mesh by key

        The mesh is defined by a strictly increasing vector of knots (edges)
        The knots can be defined from a pre-existing 2d bsplines.

        If deg is provided, immediately adds a bsplines

        Example:
        --------
                >>> from bsplines2d import BSplines2D
                >>> bs = BSplines2D()
                >>> bs.add_mesh_1d(key='m0', knots=np.linspace(0, 10, 11))

        """

        # check input data and get input dicts
        key, dref, ddata, dobj = _checks_1d.check(
            coll=self,
            # key
            key=key,
            # mesh knots
            knots=knots,
            uniform=uniform,
            # defined from pre-existing bsplines
            subkey=subkey,
            # additional attributes
            **kwdargs,
        )

        # update dicts
        self.update(dref=dref, ddata=ddata, dobj=dobj)

        # optional bspline
        if deg is not None:
            self.add_bsplines(key=key, deg=deg)

    def add_mesh_2d_rect(
        self,
        # rectangular mesh
        key=None,
        # knots
        knots0=None,
        knots1=None,
        domain=None,
        res=None,
        # cropping
        crop_poly=None,
        thresh_in=None,
        remove_isolated=None,
        # defined from pre-existing bsplines
        subkey0=None,
        subkey1=None,
        # direct addition of bsplines
        deg=None,
        **kwdargs,
    ):
        """ Add a mesh by key and domain / resolution

        Can create a 2d rectangular or triangular mesh:
            - knots0:
            - knots1:
            # or
            - domain:
            - res:


        Can optionally be cropped by a closed polygon crop_poly, that can be:
            - a (2, N) np.narray of (R, Z) coordinates
            - a tuple (Config, key_struct) to designate a struct poly

        Args thresh_in and remove_isolated control the level of cropping:
            - thresh_in:
            - remove_isolated:

        If deg is provided, immediately adds a bsplines

        """

        # get domain, poly from crop_poly
        if crop_poly is not None:
            domain, poly = _checks_2d_rect._from_croppoly(
                crop_poly=crop_poly,
                domain=domain,
            )
        else:
            poly = None

        # check input data and get input dicts
        key, dref, ddata, dobj = _checks_2d_rect.check(
            coll=self,
            # key
            key=key,
            # knots
            knots0=knots0,
            knots1=knots1,
            # rectangular
            domain=domain,
            res=res,
            # defined from pre-existing bsplines
            subkey0=subkey0,
            subkey1=subkey1,
            # attributes
            **kwdargs,
        )

        # update dicts
        self.update(dref=dref, ddata=ddata, dobj=dobj)

        # optional bspline
        if deg is not None:
            self.add_bsplines(key=key, deg=deg)

        # optional cropping
        if poly is not None:
            self.crop(
                key=key,
                crop=poly,
                thresh_in=thresh_in,
                remove_isolated=remove_isolated,
            )

    def add_mesh_2d_tri(
        self,
        # rectangular mesh
        key=None,
        # triangular mesh
        knots=None,
        indices=None,
        # from points and polygon
        pts_x0=None,
        pts_x1=None,
        # direct addition of bsplines
        deg=None,
        **kwdargs,
    ):
        """ Add a mesh by key and domain / resolution

        Can create a 2d tringular mesh:
            - knots: (nknots, 2) array of (R, Z) coordinates
            - cents: (ncents, 3 or 4) array of int indices

        If deg is provided, immediately adds a bsplines

        """

        # check input data and get input dicts
        key, dref, ddata, dobj = _checks_2d_tri.check(
            coll=self,
            # key
            key=key,
            # triangular
            knots=knots,
            indices=indices,
            # from pts and polygon
            pts_x0=pts_x0,
            pts_x1=pts_x1,
            # others
            **kwdargs,
        )

        # update dicts
        self.update(dref=dref, ddata=ddata, dobj=dobj)

        # optional bspline
        if deg is not None:
            self.add_bsplines(key=key, deg=deg)

    # def add_mesh_2d_polar(
        # self,
        # # polar mesh
        # key=None,
        # radius=None,
        # angle=None,
        # # Defined on
        # radius2d=None,
        # angle2d=None,
        # # res for contour discontinuity of angle2d
        # res=None,
        # # parameters
        # radius_dim=None,
        # radius_quant=None,
        # radius_name=None,
        # radius_units=None,
        # angle_dim=None,
        # angle_quant=None,
        # angle_name=None,
        # # direct addition of bsplines
        # deg=None,
        # **kwdargs,
    # ):
        # """ Add a 2d polar mesh

        # For now only includes radial mesh
        # radius has to be backed-up by:
            # - a radius quantity from a pre-existing rect or tri mesh
            # - a function

        # """

        # # check input data and get input dicts
        # dref, ddata, dmesh = _checks_2d_polar.check(
            # coll=self,
            # # key
            # key=key,
            # # polar
            # radius=radius,
            # angle=angle,
            # radius2d=radius2d,
            # angle2d=angle2d,
            # # parameters
            # radius_dim=radius_dim,
            # radius_quant=radius_quant,
            # radius_name=radius_name,
            # radius_units=radius_units,
            # angle_dim=angle_dim,
            # angle_quant=angle_quant,
            # angle_name=angle_name,
        # )

        # # update dicts
        # self.update(dref=dref, ddata=ddata, dobj=dobj)

        # # special treatment of angle2d
        # if dmesh[key]['angle2d'] is not None:
            # drefa, ddataa, kR, kZ = _compute.angle2d_zone(
                # coll=self,
                # key=dmesh[key]['angle2d'],
                # keyrad2d=dmesh[key]['radius2d'],
                # key_ptsO=dmesh[key]['pts_O'],
                # res=res,
                # keym0=key,
            # )

            # # update dicts
            # self.update(dref=drefa, ddata=ddataa)
            # if 'azone' in self.get_lparam(self._which_mesh):
                # self.set_param(
                    # key=key,
                    # param='azone',
                    # value=(kR, kZ),
                    # which=self._which_mesh,
                # )
            # else:
                # self.add_param(
                    # 'azone', value={key: (kR, kZ)}, which=self._which_mesh,
                # )

        # # optional bspline
        # if deg is not None:
            # self.add_bsplines(key=key, deg=deg)

    # -----------------
    # remove mesh
    # ------------------

    def remove_mesh(self, key=None, propagate=None):
        return _generic_mesh.remove_mesh(
            coll=self,
            key=key,
            propagate=propagate,
        )

    # -----------------
    # crop
    # ------------------

    def crop(self, key=None, crop=None, thresh_in=None, remove_isolated=None):
        """ Crop a mesh / bspline

        Uses:
            - a mask of bool for each mesh elements
            - a 2d (R, Z) closed polygon

        If applied on a bspline, cropping is double-checked to make sure
        all remaining bsplines have full support domain
        """
        crop, key, thresh_in = _cropping.crop(
            coll=self,
            key=key,
            crop=crop,
            thresh_in=thresh_in,
            remove_isolated=remove_isolated,
        )

        # add crop data
        keycrop = f'{key}_crop'
        ref = tuple([
            self._ddata[k0]['ref'][0]
            for k0 in self._dobj[self._which_mesh][key]['cents']
        ])
        self.add_data(
            key=keycrop,
            data=crop,
            ref=ref,
            dim='bool',
            quant='bool',
        )

        # update obj
        self._dobj[self._which_mesh][key]['crop'] = keycrop
        self._dobj[self._which_mesh][key]['crop-thresh'] = thresh_in

    # -----------------
    # indices
    # ------------------

    def select_ind(
        self,
        key=None,
        ind=None,
        elements=None,
        returnas=None,
        crop=None,
    ):
        """ Return ind for selected key (mesh or bspline) as:
                - tuple (default)
                - 'flat'

        Can covert one into the other
        """
        return _select._select_ind(
            coll=self,
            key=key,
            ind=ind,
            elements=elements,
            returnas=returnas,
            crop=crop,
        )

    def select_mesh_elements(
        self,
        key=None,
        ind=None,
        elements=None,
        returnas=None,
        return_neighbours=None,
        crop=None,
    ):
        """ Return indices or values of selected knots / cent

        Can be used to convert tuple (R, Z) indices to flat (RZ,) indices
        Can return values instead of indices
        Can return indices / values of neighbourgs

        """
        # check key
        key, _, _ = _generic_mesh._get_key_mesh_vs_bplines(
            coll=self,
            key=key,
            which=self._which_mesh,
        )

        # get ind
        mtype = self.dobj[self._which_mesh][key]['type']
        if mtype == 'rect':
            returnas_ind = tuple
        else:
            returnas_ind = bool

        ind = self.select_ind(
            key=key,
            ind=ind,
            elements=elements,
            returnas=returnas_ind,
            crop=crop,
        )

        return _select._select_mesh(
            coll=self,
            key=key,
            ind=ind,
            elements=elements,
            returnas=returnas,
            return_neighbours=return_neighbours,
        )

    # -----------------
    # interp tools
    # ------------------

    def get_sample_mesh(
        self,
        key=None,
        res=None,
        grid=None,
        mode=None,
        x0=None,
        x1=None,
        Dx0=None,
        Dx1=None,
        imshow=None,
        in_mesh=None,
        # store
        store=None,
        kx0=None,
        kx1=None,
    ):
        """ Return a sampled version of the chosen mesh """
        return _sample.sample_mesh(
            coll=self,
            key=key,
            res=res,
            grid=grid,
            mode=mode,
            x0=x0,
            x1=x1,
            Dx0=Dx0,
            Dx1=Dx1,
            imshow=imshow,
            in_mesh=in_mesh,
            # store
            store=store,
            kx0=kx0,
            kx1=kx1,
        )

    # -----------------
    # outline
    # ------------------

    def get_mesh_outline(
        self,
        key=None,
        closed=None,
    ):
        """ Get outline of 2d mesh """
        return _outline._get_outline(
            coll=self,
            key=key,
            closed=closed,
        )

    # -----------------
    # plotting
    # ------------------

    def plot_mesh(
        self,
        key=None,
        ind_knot=None,
        ind_cent=None,
        crop=None,
        color=None,
        dax=None,
        dmargin=None,
        fs=None,
        dleg=None,
        connect=None,
    ):

        return _plot.plot_mesh(
            coll=self,
            key=key,
            ind_knot=ind_knot,
            ind_cent=ind_cent,
            crop=crop,
            color=color,
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            dleg=dleg,
            connect=connect,
        )



# -*- coding: utf-8 -*-


# Common
import numpy as np
import datastock as ds


# ###############################################################
# ###############################################################
#               Main
# ###############################################################


def _get_outline(coll=None, key=None, closed=None):

    # ------------
    # check inputs

    # key
    wm = coll._which_mesh
    lok = [
        k0 for k0, v0 in coll.dobj.get(wm, {}).items()
        if v0['nd'] == '2d'
    ]
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    mtype = coll.dobj[wm][key]['type']

    # closed
    closed = ds._generic_check._check_var(
        closed, 'closed',
        types=bool,
        default=True,
    )

    # ------------
    # compute

    if mtype == 'rect':
        x0, x1 = _rect_outline(
            coll=coll,
            key=key,
        )

    else:
        x0, x1 = _tri_outline(
            coll=coll,
            key=key,
        )

    # close
    if closed is True:
        x0 = np.append(x0, x0[0])
        x1 = np.append(x1, x1[0])

    # -------------
    # format output

    k0, k1 = coll.dobj[wm][key]['knots']

    dout = {
        'x0': {
            'data': x0,
            'units': coll.ddata[k0]['units'],
            'dim': coll.ddata[k0]['dim'],
            'name': coll.ddata[k0]['name'],
            'quant': coll.ddata[k0]['quant'],
        },
        'x1': {
            'data': x1,
            'units': coll.ddata[k1]['units'],
            'dim': coll.ddata[k1]['dim'],
            'name': coll.ddata[k1]['name'],
            'quant': coll.ddata[k1]['quant'],
        },
    }

    return dout


# ###############################################################
# ###############################################################
#               rectangular
# ###############################################################


def _rect_outline(coll=None, key=None):

    # ----------
    # prepare

    wm = coll._which_mesh

    # knots
    k0, k1 = coll.dobj[wm][key]['knots']
    knots0 = coll.ddata[k0]['data']
    knots1 = coll.ddata[k1]['data']

    # crop
    crop = coll.dobj[wm][key]['crop']

    # -------
    # compute

    if crop is False:
        x0min, x0max = knots0.min(), knots0.max()
        x1min, x1max = knots1.min(), knots1.max()
        x0 = np.r_[x0min, x0max, x0max, x0min]
        x1 = np.r_[x1min, x1min, x1max, x1max]

    else:
        crop = coll.ddata[crop]['data']

        # indices of knots
        (i0, i1), (ic0, ic1) = coll.select_mesh_elements(
            key=key,
            crop=True,
            elements='knots',
            returnas='ind',
            return_neighbours=True,
        )

        # only keep edge knots
        i0min, i1min = i0.min(), i1.min()
        i0max, i1max = i0.max(), i1.max()

        # get indices of edge knots
        iok = (ic0 >= 0) & (ic1 >= 0)
        ind0 = np.array([
            np.any(~crop[ic0[ii, iok[ii, :]], ic1[ii, iok[ii, :]]])
            for ii in range(i0.size)
        ])

        # get subest of extreme coordinates
        ind = (
            ind0
            | (i0 == i0min) | (i0 == i0max)
            | (i1 == i1min) | (i1 == i1max)
        )
        if not np.any(ind):
            import pdb; pdb.set_trace()     # DB
            pass

        # keep only limits
        i0 = i0[ind]
        i1 = i1[ind]
        ic0 = ic0[ind, :]
        ic1 = ic1[ind, :]

        # get starting point
        i0_min = np.min(i0)
        i1_min = np.min(i1[i0 == i0_min])
        ii0 = ((i0 == i0_min) & (i1 == i1_min)).nonzero()[0][0]

        # pall = list(range(0, i0.size))

        p0 = [i0[ii0], i1[ii0]]
        pall = np.array([i0, i1]).T.tolist()

        lp = [p0]
        pall.remove(p0)

        old = None
        # while len(lp) == 1 or lp[-1] != p0:
        while len(pall) > 0:
            pp, old, ii0 = _next_pp(
                p0=lp[-1],
                ii0=ii0,
                pall=pall,
                i0=i0,
                i1=i1,
                ic0=ic0,
                ic1=ic1,
                crop=crop,
                old=old,
            )

            # ------ DEBUG --------
            if pp is None:
                import matplotlib.pyplot as plt
                dax = coll.plot_mesh(key)
                plt.gca().plot(
                    knots0[np.array(lp)[:, 0]],
                    knots1[np.array(lp)[:, 1]],
                    ls='-',
                    marker='o',
                    lw=2,
                )
                plt.gca().plot(
                    knots0[i0],
                    knots1[i1],
                    ls='None',
                    marker='x',
                )
                import pdb; pdb.set_trace()     # DB
            # -----------------------

            lp.append(pp)

        i0, i1 = np.array(lp).T
        x0 = knots0[i0]
        x1 = knots1[i1]

    return x0, x1


def _next_pp(
    p0=None,
    ii0=None,
    pall=None,
    i0=None,
    i1=None,
    ic0=None,
    ic1=None,
    crop=None,
    old=None,
):

    # inc0, inc1
    p1 = np.copy(p0)
    found = False
    ldir = [(1, 0), (0, 1), (-1, 0), (0, -1)]
    if old is not None:
        ldir = [pp for pp in ldir if pp != old]

    for (ip0, ip1) in ldir:

        stop = False
        while stop is False:

            p2 = [p1[0] + ip0, p1[1] + ip1]

            if p2 in pall:

                ii2 = ((i0 == p2[0]) & (i1 == p2[1])).nonzero()[0][0]
                icc0 = np.intersect1d(
                    ic0[ii0, :],
                    ic0[ii2, :],
                    assume_unique=False,
                    return_indices=False,
                )
                icc1 = np.intersect1d(
                    ic1[ii0, :],
                    ic1[ii2, :],
                    assume_unique=False,
                    return_indices=False,
                )

                i00 = np.array([
                    (ic0[ii0, :] == cc0).nonzero()[0] for cc0 in icc0
                ]).ravel()
                i11 = np.array([
                    (ic1[ii0, :] == cc1).nonzero()[0] for cc1 in icc1
                ]).ravel()

                ic = np.intersect1d(
                    i00,
                    i11,
                    assume_unique=False,
                    return_indices=False,
                )

                c0 = np.any(crop[ic0[ii0, ic], ic1[ii0, ic]])

                if c0:
                    p1 = p2
                    ii0 = ii2
                    found = True
                    pall.remove(p2)

                else:
                    stop = True

            else:
                stop = True

        if found is True:
            new = (-ip0, -ip1)
            break

    if found is False:
        return None, None, None

    return p1, new, ii0


# ###############################################################
# ###############################################################
#               triangular
# ###############################################################


def _tri_outline(coll=None, key=None):

    msg = "outline not implemented yet for triangular meshes"
    raise NotImplementedError(msg)



# -*- coding: utf-8 -*-
"""
Created on Thu Jan  5 20:14:40 2023

@author: dvezinet
"""


import warnings


import numpy as np
import datastock as ds


# ############################################################
# ############################################################
#               interpolate spectral
# ############################################################


def binning(
    coll=None,
    data=None,
    data_units=None,
    axis=None,
    # binning
    bins0=None,
    bins1=None,
    bin_data0=None,
    bin_data1=None,
    bin_units0=None,
    # kind of binning
    integrate=None,
    statistic=None,
    # options
    safety_ratio=None,
    dref_vector=None,
    ref_vector_strategy=None,
    verb=None,
    returnas=None,
    # storing
    store=None,
    store_keys=None,
):
    """ Return the spectrally interpolated coefs

    Either E xor Ebins can be provided
    - E: return interpolated coefs
    - Ebins: return binned (integrated) coefs
    """

    # ----------
    # checks

    # keys
    isbs, bin_data0 = _check_bs(
        coll=coll, 
        bin_data0=bin_data0,
        bin_data1=bin_data1,
    )

    # ----------
    # trivial

    nobin = False
    if isbs:
        
        # add ref and data
        kr, kd, ddatan, nobin = _interpolate(
            coll=coll,
            data=data,
            data_units=data_units,
            # binning
            bins0=bins0,
            bin_data0=bin_data0,
            # options
            dref_vector=dref_vector,
            verb=verb,
            store=store,
            store_keys=store_keys,
        )
        
        # safety check
        if nobin is False:
            lk = list(ddatan.keys())
            data = [ddatan[k0]['data'] for k0 in lk]
            bin_data0 = [ddatan[k0]['bin_data'] for k0 in lk]
        
    # --------------------
    # do the actua binning
    
    if nobin is False:
        dout = ds._class1_binning.binning(
            coll=coll,
            data=data,
            data_units=data_units,
            axis=axis,
            # binning
            bins0=bins0,
            bins1=bins1,
            bin_data0=bin_data0,
            bin_data1=bin_data1,
            bin_units0=bin_units0,
            # kind of binning
            integrate=integrate,
            statistic=statistic,
            # options
            safety_ratio=safety_ratio,
            dref_vector=dref_vector,
            ref_vector_strategy=ref_vector_strategy,
            verb=verb,
            returnas=True,
            # storing
            store=store,
            store_keys=store_keys,
        )

        # --------------------------------
        # remove intermediate ref and data
    
        if isbs is True:
            for dd in data + bin_data0 + [kd]:
                if dd in coll.ddata.keys():
                    coll.remove_data(dd)
            if kr in coll.dref.keys():
                coll.remove_ref(kr)
                
            for k0 in data:
                k1 = [k1 for k1, v1 in ddatan.items() if v1['data'] == k0][0]
                dout[k1] = dict(dout[k0])
                del dout[k0]
    else:
        dout = nobin

    # ----------
    # return

    if returnas is True:
        return dout


# ######################################################
# ######################################################
#                   check
# ######################################################


def _check_bs(
    coll=None,
    bin_data0=None,
    bin_data1=None,
):
    
    wbs = coll._which_bsplines
    lok_bs = [
        k0 for k0, v0 in coll.dobj.get(wbs, {}).items()
        if len(v0['ref']) == 1
    ]
    lok_dbs = [
        k0 for k0, v0 in coll.ddata.items()
        if v0.get(wbs) is not None
        and len(v0[wbs]) == 1
        and v0[wbs][0] in coll.dobj.get(wbs, {}).keys()
        and len(coll.dobj[wbs][v0[wbs][0]]['ref']) == 1
    ]
        
    c0 = (
        isinstance(bin_data0, str)
        and bin_data1 is None
        and bin_data0 in lok_dbs + lok_bs
    )
    
    if bin_data0 in lok_bs:
        bin_data0 = coll.dobj[wbs][bin_data0]['apex'][0]
    
    return c0, bin_data0
        

# ######################################################
# ######################################################
#                   interpolate
# ######################################################


def _interpolate(
    coll=None,
    data=None,
    data_units=None,
    # binning
    bins0=None,
    bin_data0=None,
    # options
    dref_vector=None,
    verb=None,
    store=None,
    store_keys=None,
):

    # ---------
    # sampling

    # mesh knots
    wm = coll._which_mesh
    wbs = coll._which_bsplines
    key_bs = coll.ddata[bin_data0][wbs][0]
    keym = coll.dobj[wbs][key_bs][wm]
    kknots = coll.dobj[wm][keym]['knots'][0]

    # resolution
    vect = coll.ddata[kknots]['data']
    res0 = np.abs(np.min(np.diff(vect)))

    # ---------
    # sampling
    
    ddata = ds._class1_binning._check_data(
        coll=coll,
        data=data,
        data_units=data_units,
        store=True,
    )
    lkdata = list(ddata.keys())
    
    # --------------------
    # bins

    dbins0 = ds._class1_binning._check_bins(
        coll=coll,
        lkdata=lkdata,
        bins=bins0,
        dref_vector=dref_vector,
        store=store,
    )

    # ----------------------
    # npts for interpolation
    
    dv = np.abs(np.diff(vect))
    dvmean = np.mean(dv) + np.std(dv)
    db = np.mean(np.diff(dbins0[lkdata[0]]['edges']))
    npts = (coll.dobj[wbs][key_bs]['deg'] + 3) * max(1, dvmean / db) + 3

    # sample mesh, update dv
    Dx0 = [dbins0[lkdata[0]]['edges'][0], dbins0[lkdata[0]]['edges'][-1]]
    xx = coll.get_sample_mesh(
        keym,
        res=res0 / npts,
        mode='abs',
        Dx0=Dx0,
    )['x0']['data']

    if xx.size == 0:
        nobins = _get_nobins(
            coll=coll,
            key_bs=key_bs,
            ddata=ddata,
            dbins0=dbins0,
            store=store,
            store_keys=store_keys,
        )
        return None, None, None, nobins

    # -------------------
    #  add ref
    
    kr = "ntemp"
    kd = "xxtemp"
    
    coll.add_ref(kr, size=xx.size)
    coll.add_data(kd, data=xx, ref=kr, units=coll.ddata[kknots]['units'])

    ddata_new = {}
    for ii, (k0, v0) in enumerate(ddata.items()):

        # interpolate bin_data
        kbdn = f"kbdn{ii}_temp"
        # try:
        coll.interpolate(
            keys=bin_data0,
            ref_key=key_bs,
            x0=kd,
            val_out=0.,
            returnas=False,
            store=True,
            inplace=True,
            store_keys=kbdn,
        )

        # except Exception as err:
        #     msg = (
        #         err.args[0]
        #         + "\n\n"
        #         f"\t- k0 = {k0}\n"
        #         f"\t- ii = {ii}\n"
        #         f"\t- bin_data0 = {bin_data0}\n"
        #         f"\t- key_bs = {key_bs}\n"
        #         f"\t- kd = {kd}\n"
        #         f"\t- xx.size: {xx.size}\n"
        #         f"\t- kbdn = {kbdn}\n"
        #     )
        #     err.args = (msg,)
        #     raise err
        
        # interpolate_data
        kdn = f"kbd{ii}_temp"
        coll.interpolate(
            keys=k0,
            ref_key=key_bs,
            x0=kd,
            val_out=0.,
            returnas=False,
            store=True,
            inplace=True,
            store_keys=kdn,
        )
        ddata_new[k0] = {'bin_data': kbdn, 'data': kdn}

    return kr, kd, ddata_new, False


def _get_nobins(
    coll=None,
    key_bs=None,
    ddata=None,
    dbins0=None,
    store=None,
    store_keys=None,
):
    
    lk = list(ddata.keys())
    wbs = coll._which_bsplines
    
    if isinstance(store_keys, str):
        store_keys = [store_keys]
    
    dout = {}
    for ii, k0 in enumerate(lk):
        
        axis = ddata[k0]['ref'].index(coll.dobj[wbs][key_bs]['ref'][0])
        
        shape = list(ddata[k0]['data'].shape)
        nb = dbins0[k0]['edges'].size - 1
        shape[axis] = nb
        
        ref = list(ddata[k0]['ref'])
        ref[axis] = dbins0[k0]['bin_ref'][0]
        
        dout[store_keys[ii]] = {
            'data': np.zeros(shape, dtype=float),
            'ref': tuple(ref),
            'units': ddata[k0]['units'],
        }
        
    if store is True:
        for k0, v0 in dout.items():
            coll.add_data(key=k0, **v0)
        
    return dout


# -*- coding: utf-8 -*-


# Built-in
import warnings


# Common
import numpy as np
import datastock as ds


# ##############################################################
# ##############################################################
#                          add data on mesh / bsplines
# ###############################################################


def add_data_meshbsplines_ref(
    coll=None,
    ref=None,
    data=None,
):

    dmesh = coll._dobj.get(coll._which_mesh)
    dbsplines = coll._dobj.get(coll._which_bsplines)

    if dmesh is None or dbsplines is None:
        return ref, data

    # ref is str
    if isinstance(ref, str):
        ref = [ref]

    # ref is tuple
    if isinstance(ref, (tuple, list)):

        # ref contains mesh
        rm = [(ii, rr) for ii, rr in enumerate(ref) if rr in dmesh.keys()]
        if len(rm) > 0:

            ref = list(ref)
            for (ki, km) in rm:
                kbs = [
                    k0 for k0, v0 in dbsplines.items()
                    if v0[coll._which_mesh] == km
                ]
                if len(kbs) == 1:
                    ref[ki] = kbs[0]
                elif len(kbs) > 1:
                    msg = (
                        "ref contains reference to mesh with several bsplines!\n"
                        f"\t- ref: {ref}\n"
                        f"\t- mesh bsplines: {kbs}\n"
                    )
                    raise Exception(msg)

        # ref contains bsplines
        rbs = [(ii, rr) for ii, rr in enumerate(ref) if rr in dbsplines.keys()]
        while len(rbs) > 0:

            ii, kb = rbs[0]

            ref = np.r_[
                ref[:ii],
                dbsplines[kb]['ref'],
                ref[ii+1:],
            ]

            rbs = [(ii, rr) for ii, rr in enumerate(ref) if rr in dbsplines.keys()]

            # repeat data if taken from ntri > 1
            data = _repeat_data_ntri(
                ref=ref,
                rbs1=kb,
                refbs=dbsplines[kb]['ref'],
                data=data,
                # mesh
                km=dbsplines[kb][coll._which_mesh],
                dmesh=dmesh,
                dbsplines=dbsplines,
            )

        ref = tuple(ref)

    return ref, data


def _repeat_data_ntri(
    ref=None,
    rbs1=None,
    refbs=None,
    data=None,
    # mesh
    km=None,
    dmesh=None,
    dbsplines=None,
):
    """ If triangular mesh with ntri > 1 => repeat data """

    c0 = (
        dmesh[km]['type'] == 'tri'
        and dmesh[km]['ntri'] > 1
    )
    if c0:
        ntri = dmesh[km]['ntri']
        indr = ref.tolist().index(refbs[0])
        nbs = dbsplines[rbs1]['shape'][0]
        ndata = data.shape[indr]
        if ndata == nbs:
            pass
        elif ndata == nbs / ntri:
            data = np.repeat(data, ntri, axis=indr)
        else:
            msg = (
                "Mismatching data shape vs multi-triangular mesh:\n"
                f"\t- data.shape[tribs] = {ndata}\n"
                f"\t- expected {nbs} / {ntri} = {nbs / ntri}\n"
            )
            raise Exception(msg)

    return data


def _set_data_bsplines(coll=None):

    if coll.dobj.get(coll._which_bsplines) is not None:

        wbs = coll._which_bsplines
        for k0, v0 in coll._ddata.items():

            lbs = [
                (v0['ref'].index(v1['ref'][0]), k1)
                for k1, v1 in coll.dobj[wbs].items()
                if v1['ref'] == tuple([
                    rr for rr in v0['ref']
                    if rr in v1['ref']
                ])
            ]

            if len(lbs) == 0:
                pass
            else:
                # re-order
                ind = np.argsort([bb[0] for bb in lbs])
                lbs = [lbs[ii][1] for ii in ind]

                # store
                coll._ddata[k0]['bsplines'] = tuple(lbs)


# ###############################################################
# ###############################################################
#                           Mesh2DRect - bsplines
# ###############################################################


def _mesh_bsplines(key=None, lkeys=None, deg=None):

    # key
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lkeys,
    )

    # deg
    deg = ds._generic_check._check_var(
        deg, 'deg',
        types=int,
        default=2,
        allowed=[0, 1, 2, 3],
    )

    # keybs
    keybs = f'{key}_bs{deg}'

    return key, keybs, deg


# ###############################################################
# ###############################################################
#                       Remove bsplines
# ###############################################################


def remove_bsplines(coll=None, key=None, propagate=None):

    # ----------
    # check

    # key
    wbs = coll._which_bsplines
    if wbs not in coll.dobj.keys():
        return

    if isinstance(key, str):
        key = [key]
    key = ds._generic_check._check_var_iter(
        key, 'key',
        types=(list, tuple),
        types_iter=str,
        allowed=coll.dobj.get(wbs, {}).keys(),
    )

    # propagate
    propagate = ds._generic_check._check_var(
        propagate, 'propagate',
        types=bool,
        default=True,
    )

    # ---------
    # remove

    for k0 in key:

        # specific data
        ldata = list(set((
            list(coll.dobj[wbs][k0]['apex'])
            + [
                k1 for k1, v1 in coll.ddata.items()
                if k0 in v1[wbs]
            ]
        )))
        for dd in ldata:
            coll.remove_data(dd, propagate=propagate)

        # specific ref
        lref = (
            coll.dobj[wbs][k0]['ref']
            + coll.dobj[wbs][k0]['ref-bs']
        )
        for rr in lref:
            if rr in coll.dref.keys():
                coll.remove_ref(rr, propagate=propagate)

        # obj
        coll.remove_obj(which=wbs, key=k0, propagate=propagate)



# -*- coding: utf-8 -*-


# Built-in
import warnings


# Common
import numpy as np
from matplotlib.tri import Triangulation as mplTri
import datastock as ds


_ELEMENTS = 'knots'


# #############################################################################
# #############################################################################
#                   names of knots and cents
# #############################################################################


def names_knots_cents(key=None, knots_name=''):

    kkr, kcr = f'{key}_nk{knots_name}', f'{key}_nc{knots_name}'
    kk, kc = f'{key}_k{knots_name}', f'{key}_c{knots_name}'

    return kkr, kcr, kk, kc


def _get_kwdargs_2d(kwdargs, latt=None):

    for k0 in latt:

        if kwdargs.get(k0) is None:
            kwdargs[k0] = [None, None]

        elif isinstance(kwdargs.get(k0), str):
            kwdargs[k0] = [kwdargs[k0], kwdargs[k0]]

        elif not (isinstance(kwdargs[k0], list) and len(kwdargs[k0]) == 2):
            msg = (
                f"Wrong attributes for 2d rect mesh '{k0}':\n"
                "Please provide units, dim, quant, name as list of len() = 2"
                f"Provided:\n{kwdargs}"
            )
            raise Exception(msg)

    dim, quant, name, units = [kwdargs[ss] for ss in latt]

    return dim, quant, name, units


# ###############################################################
# ###############################################################
#               mesh vs bsplines
# ###############################################################


def _get_key_mesh_vs_bplines(
    coll=None,
    key=None,
    which=None,
):

    wm = coll._which_mesh
    wbs = coll._which_bsplines
    if which in [None, wm]:
        lk1 = list(coll.dobj.get(wm, {}).keys())
    else:
        lk1 = []

    if which in [None, wbs]:
        lk2 = list(coll.dobj.get(wbs, {}).keys())
    else:
        lk2 = []

    # key
    key = ds._generic_check._check_var(
        key, 'key',
        allowed=lk1 + lk2,
        types=str,
    )

    # which
    if key in lk1:
        cat = wm
    else:
        cat = wbs
    assert which in [cat, None], (cat, which)

    # keys
    if cat == wbs:
        keym = coll.dobj[wbs][key][wm]
        keybs = key
    else:
        keym = key
        keybs = None

    return keym, keybs, cat


# #############################################################################
# #############################################################################
#                           Mesh2DRect - select
# #############################################################################


def _select_ind_check(
    ind=None,
    elements=None,
    returnas=None,
    crop=None,
    shape2d=None,
):

    # ----------------------
    # check basic conditions

    if shape2d:
        lc = [
            ind is None,
            isinstance(ind, tuple)
            and len(ind) == 2
            and (
                all([np.isscalar(ss) for ss in ind])
                or all([
                    hasattr(ss, '__iter__')
                    and len(ss) == len(ind[0])
                    for ss in ind
                ])
                or all([isinstance(ss, np.ndarray) for ss in ind])
            ),
            (
                np.isscalar(ind)
                or (
                    hasattr(ind, '__iter__')
                    and all([np.isscalar(ss) for ss in ind])
                )
                or isinstance(ind, np.ndarray)
            )
        ]

    else:
        lc = [
            ind is None,
            np.isscalar(ind)
            or (
                hasattr(ind, '__iter__')
                and all([np.isscalar(ss) for ss in ind])
            )
            or isinstance(ind, np.ndarray)
        ]

    # check lc
    if not any(lc):
        if shape2d:
            msg = (
                "Arg ind must be either:\n"
                "\t- None\n"
                "\t- int or array of int: int indices in mixed (R, Z) index\n"
                "\t- tuple of such: int indices in (R, Z) index respectively\n"
                f"Provided: {ind}"
            )
        else:
            msg = (
                "Arg ind must be either:\n"
                "\t- None\n"
                "\t- int or array of int: int indices\n"
                "\t- array of bool: bool indices\n"
                f"Provided: {ind}"
            )
        raise Exception(msg)

    # ----------------------
    # adapt to each case

    if lc[0]:
        pass

    elif lc[1] and shape2d:
        if any([not isinstance(ss, np.ndarray) for ss in ind]):
            ind = (
                np.atleast_1d(ind[0]).astype(int),
                np.atleast_1d(ind[1]).astype(int),
            )
        lc0 = [
            [
                isinstance(ss, np.ndarray),
                np.issubdtype(ss.dtype, np.integer),
                ss.shape == ind[0].shape,
            ]
                for ss in ind
        ]
        if not all([all(cc) for cc in lc0]):
            ltype = [type(ss) for ss in ind]
            ltypes = [
                ss.dtype if isinstance(ss, np.ndarray) else False
                for ss in ind
            ]
            lshapes = [
                ss.shape if isinstance(ss, np.ndarray) else len(ss)
                for ss in ind
            ]
            msg = (
                "Arg ind must be a tuple of 2 arrays of int of same shape\n"
                f"\t- lc0: {lc0}\n"
                f"\t- types: {ltype}\n"
                f"\t- type each: {ltypes}\n"
                f"\t- shape: {lshapes}\n"
                f"\t- ind: {ind}"
            )
            raise Exception(msg)

    elif lc[1] and not shape2d:
        if not isinstance(ind, np.ndarray):
            ind = np.atleast_1d(ind).astype(int)
        c0 = (
            np.issubdtype(ind.dtype, np.integer)
            or np.issubdtype(ind.dtype, np.bool_)
        )
        if not c0:
            msg = (
                "Arg ind must be an array of bool or int\n"
                f"Provided: {ind.dtype}"
            )
            raise Exception(msg)

    else:
        if not isinstance(ind, np.ndarray):
             ind = np.atleast_1d(ind).astype(int)
        c0 = (
            np.issubdtype(ind.dtype, np.integer)
            or np.issubdtype(ind.dtype, np.bool_)
        )
        if not c0:
            msg = (
                 "Arg ind must be an array of bool or int\n"
                 f"Provided: {ind.dtype}"
            )
            raise Exception(msg)

    # elements
    elements = ds._generic_check._check_var(
        elements, 'elements',
        types=str,
        default=_ELEMENTS,
        allowed=['knots', 'cents'],
    )

    # returnas
    if shape2d:
        retdef = tuple
        retok = [tuple, np.ndarray, 'tuple-flat', 'array-flat', bool]
    else:
        retdef = bool
        retok = [int, bool]

    returnas = ds._generic_check._check_var(
        returnas, 'returnas',
        types=None,
        default=retdef,
        allowed=retok,
    )

    # crop
    crop = ds._generic_check._check_var(
        crop, 'crop',
        types=bool,
        default=True,
    )

    return ind, elements, returnas, crop


def _select_check(
    elements=None,
    returnas=None,
    return_ind_as=None,
    return_neighbours=None,
):

    # elements
    elements = ds._generic_check._check_var(
        elements, 'elements',
        types=str,
        default=_ELEMENTS,
        allowed=['knots', 'cents'],
    )

    # returnas
    returnas = ds._generic_check._check_var(
        returnas, 'returnas',
        types=None,
        default='ind',
        allowed=['ind', 'data'],
    )

    # return_ind_as
    return_ind_as = ds._generic_check._check_var(
        return_ind_as, 'return_ind_as',
        types=None,
        default=int,
        allowed=[int, bool],
    )

    # return_neighbours
    return_neighbours = ds._generic_check._check_var(
        return_neighbours, 'return_neighbours',
        types=bool,
        default=True,
    )

    return elements, returnas, return_ind_as, return_neighbours,


# #############################################################################
# #############################################################################
#                           Mesh2DRect - bsplines
# #############################################################################


def _mesh_bsplines(key=None, lkeys=None, deg=None):

    # key
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lkeys,
    )

    # deg
    deg = ds._generic_check._check_var(
        deg, 'deg',
        types=int,
        default=2,
        allowed=[0, 1, 2, 3],
    )

    # keybs
    keybs = f'{key}-bs{deg}'

    return key, keybs, deg


# ###############################################################
# ###############################################################
#                       Remove bsplines
# ###############################################################


def remove_mesh(coll=None, key=None, propagate=None):

    # ----------
    # check

    # key
    wm = coll._which_mesh
    wbs = coll._which_bsplines

    if wm not in coll.dobj.keys():
        return

    if isinstance(key, str):
        key = [key]
    key = ds._generic_check._check_var_iter(
        key, 'key',
        types=(list, tuple),
        types_iter=str,
        allowed=coll.dobj.get(wm, {}).keys(),
    )

    # propagate
    propagate = ds._generic_check._check_var(
        propagate, 'propagate',
        types=bool,
        default=True,
    )

    # ---------
    # remove

    for k0 in key:

        # remove bsplines
        if wbs in coll.dobj.keys():
            lbs = [k1 for k1, v1 in coll.dobj[wbs].items() if v1[wm] == k0]
            coll.remove_bsplines(key=lbs, propagate=propagate)

        # specific data
        ldata = list(
            coll.dobj[wm][k0]['knots']
            + coll.dobj[wm][k0]['cents']
        )
        lref = [coll.ddata[kk]['ref'] for kk in ldata]
        crop = coll.dobj[wm][k0]['crop']
        if crop is not None and crop is not False:
            ldata.append(crop)

        for dd in ldata:
            coll.remove_data(dd, propagate=propagate)

        # specific ref
        for rr in lref:
            if rr in coll.dref.keys():
                coll.remove_ref(rr, propagate=propagate)

        # obj
        coll.remove_obj(which=wm, key=k0, propagate=propagate)



# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import datastock as ds

# specific
from . import _generic_check
from . import _class01_checks as _checks


# ################################################################
# ################################################################
#                     Main plotting
# ################################################################


def plot_bspline(
    # ressources
    coll=None,
    # inputs
    key=None,
    indbs=None,
    indt=None,
    # parameters
    knots=None,
    cents=None,
    res=None,
    plot_mesh=None,
    val_out=None,
    nan0=None,
    # plot-specific
    cmap=None,
    dax=None,
    dmargin=None,
    fs=None,
    dleg=None,
):

    # --------------
    # check input

    (
        key, keym0, keym, mtype0, mtype,
        indbs, indt,
        knots, cents, knotsi, centsi,
        plot_mesh, cmap, dleg,
    ) = _plot_bspline_check(
        coll=coll,
        key=key,
        indbs=indbs,
        indt=indt,
        knots=knots,
        cents=cents,
        plot_mesh=plot_mesh,
        cmap=cmap,
        dleg=dleg,
    )

    # --------------
    #  Prepare data

    bspline, extent, interp = _plot_bspline_prepare(
        coll=coll,
        key=key,
        keym=keym,
        mtype0=mtype0,
        mtype=mtype,
        indbs=indbs,
        indt=indt,
        knotsi=knotsi,
        centsi=centsi,
        res=res,
        val_out=val_out,
        nan0=nan0,
    )

    # --------------
    # plot - prepare

    if dax is None:

        if dmargin is None:
            dmargin = {
                'left': 0.1, 'right': 0.9,
                'bottom': 0.1, 'top': 0.9,
                'hspace': 0.1, 'wspace': 0.1,
            }

        fig = plt.figure(figsize=fs)
        gs = gridspec.GridSpec(ncols=1, nrows=1, **dmargin)
        ax0 = fig.add_subplot(gs[0, 0], aspect='equal')
        ax0.set_xlabel(f'R (m)')
        ax0.set_ylabel(f'Z (m)')

        dax = {'cross': ax0}

    dax = _generic_check._check_dax(dax=dax, main='cross')

    # --------------
    # plot

    if plot_mesh is True:
        keym = coll.dobj['bsplines'][key]['mesh']
        if mtype0 == 'polar':
            _ = coll.plot_mesh(key=keym, dleg=False)
        else:
            dax = coll.plot_mesh(key=keym, dax=dax, dleg=False)

    kax = 'cross'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        ax.imshow(
            bspline,
            extent=extent,
            interpolation=interp,
            origin='lower',
            aspect='equal',
            cmap=cmap,
            vmin=0.,
            vmax=1.,
        )

        if mtype0 != 'polar':
            if knots is not False:
                ax.plot(
                    knotsi[0].ravel(),
                    knotsi[1].ravel(),
                    marker='x',
                    ms=6,
                    ls='None',
                    color='k',
                )

            if cents is not False:
                ax.plot(
                    centsi[0].ravel(),
                    centsi[1].ravel(),
                    marker='o',
                    ms=6,
                    ls='None',
                    color='k',
                )

        ax.relim()
        ax.autoscale()

        # --------------
        # dleg

        if dleg is not False:
            ax.legend(**dleg)

    return dax


# ############################################################
# ############################################################
#                           plot bspline
# ############################################################


def _plot_bsplines_get_dRdZ(coll=None, km=None, meshtype=None):
    # Get minimum distances

    if meshtype == 'rect':
        kR, kZ = coll.dobj['mesh'][km]['knots']
        Rk = coll.ddata[kR]['data']
        Zk = coll.ddata[kZ]['data']
        dR = np.min(np.diff(Rk))
        dZ = np.min(np.diff(Zk))

    elif meshtype == 'tri':
        indtri = coll.ddata[coll.dobj['mesh'][km]['ind']]['data']
        kknots = coll.dobj['mesh'][km]['knots']
        Rk = coll.ddata[kknots[0]]['data']
        Zk = coll.ddata[kknots[1]]['data']
        R = Rk[indtri]
        Z = Zk[indtri]
        dist = np.mean(np.array([
            np.sqrt((R[:, 1] - R[:, 0])**2 + (Z[:, 1] - Z[:, 0])**2),
            np.sqrt((R[:, 2] - R[:, 1])**2 + (Z[:, 2] - Z[:, 1])**2),
            np.sqrt((R[:, 2] - R[:, 0])**2 + (Z[:, 2] - Z[:, 0])**2),
        ]))
        dR, dZ = dist, dist

    else:
        km2 = coll.dobj[coll._which_mesh][km]['submesh']
        meshtype = coll.dobj[coll._which_mesh][km2]['type']
        return _plot_bsplines_get_dRdZ(
            coll=coll, km=km2, meshtype=meshtype,
        )

    Rminmax = [Rk.min(), Rk.max()]
    Zminmax = [Zk.min(), Zk.max()]
    return dR, dZ, Rminmax, Zminmax


# ###############################################################
# ###############################################################
#                           checks
# ###############################################################


def _plot_bspline_check(
    coll=None,
    key=None,
    indbs=None,
    indt=None,
    knots=None,
    cents=None,
    plot_mesh=None,
    cmap=None,
    dleg=None,
):

    # key
    (
     which_mesh, which_bsplines, keym, key, cat,
     ) = _checks._get_key_mesh_vs_bplines(
        coll=coll,
        key=key,
        forcecat='bsplines',
    )

    keym0 = coll.dobj[which_bsplines][key][which_mesh]
    mtype0 = coll.dobj[which_mesh][keym0]['type']
    if mtype0 == 'polar':
        keym = coll.dobj[which_mesh][keym0]['submesh']
        mtype = coll.dobj[which_mesh][keym]['type']
    else:
        keym = keym0
        mtype = mtype0

    # knots, cents
    knots = ds._generic_check._check_var(
        knots, 'knots', default=True, types=bool,
    )
    cents = ds._generic_check._check_var(
        cents, 'cents', default=True, types=bool,
    )

    # ind_bspline
    if indbs is not None:
        indbs = coll.select_bsplines(
            key=key,
            ind=indbs,
            returnas='ind',
            return_knots=False,
            return_cents=False,
            crop=False,
        )

    _, knotsi, centsi = coll.select_bsplines(
        key=key,
        ind=indbs,
        returnas='data',
        return_knots=True,
        return_cents=True,
        crop=False,
    )

    # indt
    nt = False
    if mtype0 == 'polar':
        radius2d = coll.dobj[which_mesh][keym0]['radius2d']
        r2d_reft = coll.get_time(key=radius2d)[2]
        if r2d_reft is not None:
            nt = coll.dref[r2d_reft]['size']

    if nt is False:
        indt = None
    else:
        if indt is None:
            indt = 0
        indt = np.atleast_1d(indt).ravel()[0]

    # plot_mesh
    plot_mesh = ds._generic_check._check_var(
        plot_mesh, 'plot_mesh',
        default=True,
        types=bool,
    )

    # cmap
    if cmap is None:
        cmap = 'viridis'

    # dleg
    defdleg = {
        'bbox_to_anchor': (1.1, 1.),
        'loc': 'upper left',
        'frameon': True,
    }
    dleg = ds._generic_check._check_var(
        dleg, 'dleg',
        default=defdleg,
        types=(bool, dict),
    )

    return (
        key, keym0, keym, mtype0, mtype,
        indbs, indt,
        knots, cents, knotsi, centsi,
        plot_mesh, cmap, dleg,
    )


# ###############################################################
# ###############################################################
#                           prepare
# ###############################################################


def _plot_bspline_prepare(
    coll=None,
    # keys
    key=None,
    keym0=None,
    keym=None,
    mtype0=None,
    mtype=None,
    # indices
    indbs=None,
    indt=None,
    # options
    res=None,
    knotsi=None,
    centsi=None,
    val_out=None,
    nan0=None,
):

    # check input
    deg = coll.dobj['bsplines'][key]['deg']

    # get dR, dZ
    dR, dZ, _, _ = _plot_bsplines_get_dRdZ(
        coll=coll, km=keym, meshtype=mtype,
    )

    # resolution of sampling
    if res is None:
        if mtype == 'rect':
            res_coef = 0.05
        else:
            res_coef = 0.25
        res = [res_coef*dR, res_coef*dZ]

    # sampling domain
    if mtype0 == 'polar':
        DR = None
        DZ = None
    else:
        knotsiR, knotsiZ = knotsi
        DR = [np.nanmin(knotsiR) + dR*1.e-10, np.nanmax(knotsiR) - dR*1.e-10]
        DZ = [np.nanmin(knotsiZ) + dZ*1.e-10, np.nanmax(knotsiZ) - dZ*1.e-10]

    # sample
    R, Z = coll.get_sample_mesh(
        key=keym,
        res=res,
        DR=DR,
        DZ=DZ,
        mode='abs', grid=True, imshow=True,
    )

    # bspline
    bspline = coll.interpolate_profile2d(
        key=key,
        R=R,
        Z=Z,
        # coefs=coefs,
        indt=indt,
        indbs=indbs,
        details=indbs is not None,
        grid=False,
        nan0=nan0,
        val_out=val_out,
        return_params=False,
    )[0]

    if indbs is None:
        if bspline.ndim == R.ndim + 1:
            assert bspline.shape[1:] == R.shape
            bspline = bspline[0, ...]
    else:
        if bspline.ndim == R.ndim + 1:
            assert bspline.shape[:-1] == R.shape
            bspline = np.nansum(bspline, axis=-1)
        elif bspline.ndim == R.ndim + 2:
            assert bspline.shape[1:-1] == R.shape
            bspline = np.nansum(bspline[0, ...], axis=-1)

    if bspline.shape != R.shape:
        import pdb; pdb.set_trace() # DB
        pass

    # extent
    if mtype0 == 'polar':
        extent = (
            R.min(), R.max(),
            Z.min(), Z.max(),
        )
    else:
        extent = (
            DR[0], DR[1],
            DZ[0], DZ[1],
        )

    # interpolation
    if deg == 0:
        interp = 'nearest'
    elif deg == 1:
        interp = 'bilinear'
    elif deg >= 2:
        interp = 'bicubic'

    return bspline, extent, interp
